"""Data types and operations from Strategic Coordination Test Data Injection 0.0.1 OpenAPI"""

# This file is autogenerated; do not modify manually!

from __future__ import annotations

from enum import Enum
from typing import Dict, List, Optional

from uas_standards import Operation

from implicitdict import ImplicitDict, StringBasedDateTime


class StatusResponseStatus(str, Enum):
    """The status of the USS automated testing interface.
    - `Starting`: the USS is starting and the automated test driver should wait before sending requests.
    - `Ready`: the USS is ready to receive test requests.
    """

    Starting = "Starting"
    Ready = "Ready"


class StatusResponse(ImplicitDict):
    status: StatusResponseStatus
    """The status of the USS automated testing interface.
    - `Starting`: the USS is starting and the automated test driver should wait before sending requests.
    - `Ready`: the USS is ready to receive test requests.
    """

    version: Optional[str]
    """Arbitrary string representing the version of the USS system to be tested."""


UUIDv4Format = str
"""String whose format matches a version-4 UUID according to RFC 4122."""


EntityID = UUIDv4Format


class TimeFormat(str, Enum):
    RFC3339 = "RFC3339"


class Time(ImplicitDict):
    value: StringBasedDateTime
    """RFC3339-formatted time/date string.  The time zone must be 'Z'."""

    format: TimeFormat = TimeFormat.RFC3339


class RadiusUnits(str, Enum):
    """FIXM-compatible units.  Only meters ("M") are acceptable for UTM."""

    M = "M"


class Radius(ImplicitDict):
    value: float
    """Distance from the centerpoint of a circular area, along the WGS84 ellipsoid."""

    units: RadiusUnits = RadiusUnits.M
    """FIXM-compatible units.  Only meters ("M") are acceptable for UTM."""


class AltitudeReference(str, Enum):
    """A code indicating the reference for a vertical distance. See AIXM 5.1 and FIXM 4.2.0. Currently, UTM only allows WGS84 with no immediate plans to allow other options. FIXM and AIXM allow for 'SFC' which is equivalent to AGL."""

    W84 = "W84"


class AltitudeUnits(str, Enum):
    """The reference quantities used to express the value of altitude. See FIXM 4.2. Currently, UTM only allows meters with no immediate plans to allow other options."""

    M = "M"


class Altitude(ImplicitDict):
    value: float
    """The numeric value of the altitude. Note that min and max values are added as a sanity check. As use cases evolve and more options are made available in terms of units of measure or reference systems, these bounds may be re-evaluated."""

    reference: AltitudeReference = AltitudeReference.W84
    """A code indicating the reference for a vertical distance. See AIXM 5.1 and FIXM 4.2.0. Currently, UTM only allows WGS84 with no immediate plans to allow other options. FIXM and AIXM allow for 'SFC' which is equivalent to AGL."""

    units: AltitudeUnits = AltitudeUnits.M
    """The reference quantities used to express the value of altitude. See FIXM 4.2. Currently, UTM only allows meters with no immediate plans to allow other options."""


Latitude = float
"""Degrees of latitude north of the equator, with reference to the WGS84 ellipsoid."""


Longitude = float
"""Degrees of longitude east of the Prime Meridian, with reference to the WGS84 ellipsoid."""


class LatLngPoint(ImplicitDict):
    """Point on the earth's surface."""

    lng: Longitude

    lat: Latitude


class Circle(ImplicitDict):
    """A circular area on the surface of the earth."""

    center: Optional[LatLngPoint]

    radius: Optional[Radius]


class OperationalIntentState(str, Enum):
    """State of an operational intent. 'Accepted': Operational intent is created and shared, but not yet in use; see standard text for more details. The create or update request for this operational intent reference must include a Key containing all OVNs for all relevant Entities. 'Activated': Operational intent is in active use; see standard text for more details. The create or update request for this operational intent reference must include a Key containing all OVNs for all relevant Entities. 'Nonconforming': UA is temporarily outside its volumes, but the situation is expected to be recoverable; see standard text for more details. In this state, the `/uss/v1/operational_intents/{entityid}/telemetry` USS-USS endpoint should respond, if available, to queries from USS peers.  The create or update request for this operational intent may omit a Key in this case because the operational intent is being adjusted as flown and cannot necessarily deconflict. 'Contingent': UA is considered unrecoverably unable to conform with its coordinate operational intent; see standard text for more details. This state must transition to Ended.  In this state, the `/uss/v1/operational_intents/{entityid}/telemetry` USS-USS endpoint should respond, if available, to queries from USS peers.  The create or update request for this operational intent may omit a Key in this case because the operational intent is being adjusted as flown and cannot necessarily deconflict."""

    Accepted = "Accepted"
    Activated = "Activated"
    Nonconforming = "Nonconforming"
    Contingent = "Contingent"


Priority = int
"""Ordinal priority of the operational intent, as defined by the regulator.  Operational intents with lesser values are lower priority than all operational intents with greater values.  A lower-priority operational intent may not create a conflict with a higher-priority operational intent.  A higher-priority operational intent may create a conflict with a lower-priority operational intent.  The regulator specifies whether an operational intent may create a conflict with other operational intents of the same priority."""


class FlightAuthorisationDataOperation_category(str, Enum):
    """Category of UAS operation (‘open’, ‘specific’, ‘certified’) as defined in COMMISSION DELEGATED REGULATION (EU) 2019/945. Required by ANNEX IV of COMMISSION IMPLEMENTING REGULATION (EU) 2021/664, paragraph 4."""

    Unknown = "Unknown"
    Open = "Open"
    Specific = "Specific"
    Certified = "Certified"


class OperationMode(str, Enum):
    """Specify if the operation is a `VLOS` or `BVLOS` operation. Required by ANNEX IV of COMMISSION IMPLEMENTING REGULATION (EU) 2021/664, paragraph 2."""

    Undeclared = "Undeclared"
    Vlos = "Vlos"
    Bvlos = "Bvlos"


class UASClass(str, Enum):
    """Specify the class of the UAS to be flown, the specifition matches EASA class identification label categories. UAS aircraft class as defined in COMMISSION DELEGATED REGULATION (EU) 2019/945 (C0 to C4) and COMMISSION DELEGATED REGULATION (EU) 2020/1058 (C5 and C6). This field is required by ANNEX IV of COMMISSION IMPLEMENTING REGULATION (EU) 2021/664, paragraph 4."""

    Other = "Other"
    C0 = "C0"
    C1 = "C1"
    C2 = "C2"
    C3 = "C3"
    C4 = "C4"
    C5 = "C5"
    C6 = "C6"


class InjectFlightResponseResult(str, Enum):
    """The result of the flight submission

      - `Planned`: The flight submission data was valid and the flight was successfully processed by the USS and is now authorized.

      - `ReadyToFly`: The flight is ready for the operator to begin flying.

      - `Rejected`: The flight submission data provided was invalid and/or could not be used to attempt to authorize the flight.

      - `ConflictWithFlight`: The flight submission data was valid, but the flight could not be authorized because of a disallowed conflict with another flight.

      - `Failed`: The USS was not able to successfully authorize the flight due to a problem with the USS or a downstream system
    """

    Planned = "Planned"
    ReadyToFly = "ReadyToFly"
    Rejected = "Rejected"
    ConflictWithFlight = "ConflictWithFlight"
    Failed = "Failed"


class InjectFlightResponse(ImplicitDict):
    result: InjectFlightResponseResult
    """The result of the flight submission

      - `Planned`: The flight submission data was valid and the flight was successfully processed by the USS and is now authorized.

      - `ReadyToFly`: The flight is ready for the operator to begin flying.

      - `Rejected`: The flight submission data provided was invalid and/or could not be used to attempt to authorize the flight.

      - `ConflictWithFlight`: The flight submission data was valid, but the flight could not be authorized because of a disallowed conflict with another flight.

      - `Failed`: The USS was not able to successfully authorize the flight due to a problem with the USS or a downstream system
    """

    notes: Optional[str]
    """Human-readable explanation of the observed result.  This explanation should be available to a human reviewing the test results, and ideally should explain why an undesirable result was obtained.  For instance, if the injection attempt Failed, then these notes may indicate that the attempt failed because the DSS indicated 400 to a valid request (perhaps also including the valid request as proof)."""

    operational_intent_id: Optional[EntityID]
    """The id of the operational intent communicated to the DSS. This value is only required when the result of the flight submission is `Planned`."""


class DeleteFlightResponseResult(str, Enum):
    """The result of attempted flight cancellation/closure

      - `Closed`: The flight was closed successfully by the USS and is now out of the UTM system.

      - `Failed`: The flight could not be closed successfully by the USS.
    """

    Closed = "Closed"
    Failed = "Failed"


class DeleteFlightResponse(ImplicitDict):
    result: DeleteFlightResponseResult
    """The result of attempted flight cancellation/closure

      - `Closed`: The flight was closed successfully by the USS and is now out of the UTM system.

      - `Failed`: The flight could not be closed successfully by the USS.
    """

    notes: Optional[str]
    """Human-readable explanation of the observed result."""


class ClearAreaOutcome(ImplicitDict):
    success: Optional[bool] = False
    """True if, and only if, all flights in the specified area owned by the USS were canceled and removed."""

    message: Optional[str]
    """If the USS was unable to clear the entire area, this message can provide information on the problem encountered."""

    timestamp: str
    """The time at which this operation was performed by the USS."""


class ClearAreaResponse(ImplicitDict):
    outcome: ClearAreaOutcome


class Capability(str, Enum):
    """Capability of a USS.

      `FlightAuthorisationValidation`: USS supports EU flight authorisation
        parameter validation.

      `BasicStrategicConflictDetection`: USS supports strategic conflict
        detection for typical flights, including future planning (Accepted
        operational intents), activation (Accepted operational intents), and
        closing (deleting the operational intent reference).

      `HighPriorityFlights`: USS supports flights at priority levels higher
        than typical flights.
    """

    FlightAuthorisationValidation = "FlightAuthorisationValidation"
    BasicStrategicConflictDetection = "BasicStrategicConflictDetection"
    HighPriorityFlights = "HighPriorityFlights"


class CapabilitiesResponse(ImplicitDict):
    capabilities: Optional[List[Capability]] = []
    """Set of capabilities supported by this USS."""


class Polygon(ImplicitDict):
    """An enclosed area on the earth. The bounding edges of this polygon are defined to be the shortest paths between connected vertices.  This means, for instance, that the edge between two points both defined at a particular latitude is not generally contained at that latitude. The winding order must be interpreted as the order which produces the smaller area. The path between two vertices is defined to be the shortest possible path between those vertices. Edges may not cross. Vertices may not be duplicated.  In particular, the final polygon vertex must not be identical to the first vertex."""

    vertices: List[LatLngPoint]


class Volume3D(ImplicitDict):
    """A three-dimensional geographic volume consisting of a vertically-extruded shape. Exactly one outline must be specified."""

    outline_circle: Optional[Circle]
    """A circular geographic shape on the surface of the earth."""

    outline_polygon: Optional[Polygon]
    """A polygonal geographic shape on the surface of the earth."""

    altitude_lower: Optional[Altitude]
    """Minimum bounding altitude of this volume. Must be less than altitude_upper, if specified."""

    altitude_upper: Optional[Altitude]
    """Maximum bounding altitude of this volume. Must be greater than altitude_lower, if specified."""


class Volume4D(ImplicitDict):
    """Contiguous block of geographic spacetime."""

    volume: Volume3D

    time_start: Optional[Time]
    """Beginning time of this volume. Must be before time_end."""

    time_end: Optional[Time]
    """End time of this volume. Must be after time_start."""


class OperationalIntentTestInjection(ImplicitDict):
    """Parameters that define an operational intent: this injection is used to create a operational intent reference in the DSS and also responding to requests for details of that operational intent (by other USSes or the test driver). The USS under test will need to process this data to both create a valid operational intent reference and responding to a query for details."""

    state: OperationalIntentState

    priority: Priority

    volumes: List[Volume4D]
    """Nominal volumes, as would be reported by a USS's operational_intents endpoint."""

    off_nominal_volumes: List[Volume4D]
    """Off-Nominal volumes, as would be reported by a USS's operational_intents endpoint."""


class FlightAuthorisationData(ImplicitDict):
    """A dataset to hold details of a UAS flight authorization request. Full description of a flight authorisation including mandatory information required by ANNEX IV of COMMISSION IMPLEMENTING REGULATION (EU) 2021/664 for an UAS flight authorisation request. Reference: https://eur-lex.europa.eu/legal-content/EN/TXT/HTML/?uri=CELEX:32021R0664&from=EN#d1e32-178-1"""

    uas_serial_number: str
    """Unique serial number of the unmanned aircraft or, if the unmanned aircraft is privately built, the unique serial number of the add-on. This is expressed in the ANSI/CTA-2063 Physical Serial Number format. Required by ANNEX IV of COMMISSION IMPLEMENTING REGULATION (EU) 2021/664, paragraph 1."""

    operation_mode: OperationMode

    operation_category: FlightAuthorisationDataOperation_category
    """Category of UAS operation (‘open’, ‘specific’, ‘certified’) as defined in COMMISSION DELEGATED REGULATION (EU) 2019/945. Required by ANNEX IV of COMMISSION IMPLEMENTING REGULATION (EU) 2021/664, paragraph 4."""

    uas_class: UASClass

    identification_technologies: List[str]
    """Technology used to identify the UAS. Required by ANNEX IV of COMMISSION IMPLEMENTING REGULATION (EU) 2021/664, paragraph 6."""

    uas_type_certificate: Optional[str]
    """Provisional field. Not applicable as of September 2021. Required only if `uas_class` is set to `other` by ANNEX IV of COMMISSION IMPLEMENTING REGULATION (EU) 2021/664, paragraph 4."""

    connectivity_methods: List[str]
    """Connectivity methods. Required by ANNEX IV of COMMISSION IMPLEMENTING REGULATION (EU) 2021/664, paragraph 7."""

    endurance_minutes: int
    """Endurance of the UAS. This is expressed in minutes. Required by ANNEX IV of COMMISSION IMPLEMENTING REGULATION (EU) 2021/664, paragraph 8."""

    emergency_procedure_url: str
    """The URL at which the applicable emergency procedure in case of a loss of command and control link may be retrieved. Required by ANNEX IV of COMMISSION IMPLEMENTING REGULATION (EU) 2021/664, paragraph 9."""

    operator_id: str
    """Registration number of the UAS operator.
    The format is defined in EASA Easy Access Rules for Unmanned Aircraft Systems GM1 to AMC1
    Article 14(6) Registration of UAS operators and ‘certified’ UAS.
    Required by ANNEX IV of COMMISSION IMPLEMENTING REGULATION (EU) 2021/664, paragraph 10.
    """

    uas_id: Optional[str]
    """When applicable, the registration number of the unmanned aircraft.
    This is expressed using the nationality and registration mark of the unmanned aircraft in
    line with ICAO Annex 7.
    Specified by ANNEX IV of COMMISSION IMPLEMENTING REGULATION (EU) 2021/664, paragraph 10.
    """


class InjectFlightRequest(ImplicitDict):
    operational_intent: Optional[OperationalIntentTestInjection]

    flight_authorisation: Optional[FlightAuthorisationData]


class ClearAreaRequest(ImplicitDict):
    request_id: str
    """Unique string identifying this request.  If a second request with an identical ID is received, the USS may return the same response from the previous operation rather than attempting to clear the area again (the USS may also attempt to clear the area again)."""

    extent: Volume4D
    """The USS should cancel and remove any flight where any part of that flight intersects this area."""


class OperationID(str, Enum):
    GetStatus = "getStatus"
    GetCapabilities = "getCapabilities"
    InjectFlight = "injectFlight"
    DeleteFlight = "deleteFlight"
    ClearArea = "clearArea"


OPERATIONS: Dict[OperationID, Operation] = {
    OperationID.GetStatus: Operation(
        id="getStatus",
        path="/v1/status",
        verb="GET",
        request_body_type=None,
        response_body_type={
            200: StatusResponse,
            401: None,
            403: None,
            404: None,
        }
    ),
    OperationID.GetCapabilities: Operation(
        id="getCapabilities",
        path="/v1/capabilities",
        verb="GET",
        request_body_type=None,
        response_body_type={
            200: CapabilitiesResponse,
            401: None,
            403: None,
        }
    ),
    OperationID.InjectFlight: Operation(
        id="injectFlight",
        path="/v1/flights/{flight_id}",
        verb="PUT",
        request_body_type=InjectFlightRequest,
        response_body_type={
            200: InjectFlightResponse,
            401: None,
            403: None,
        }
    ),
    OperationID.DeleteFlight: Operation(
        id="deleteFlight",
        path="/v1/flights/{flight_id}",
        verb="DELETE",
        request_body_type=None,
        response_body_type={
            200: DeleteFlightResponse,
            401: None,
            403: None,
        }
    ),
    OperationID.ClearArea: Operation(
        id="clearArea",
        path="/v1/clear_area_requests",
        verb="POST",
        request_body_type=ClearAreaRequest,
        response_body_type={
            200: ClearAreaResponse,
            401: None,
            403: None,
        }
    ),
}



import keyword
from typing import List

from data_types import DataType, ObjectField, is_primitive_python_type
from operations import Operation


def escape_escapes(s: str) -> str:
    return s.replace("\\", "\\\\")


def docstring_comment(lines: List[str]) -> List[str]:
    if len(lines) == 0:
        return []
    if len(lines) == 1:
        return [f'"""{escape_escapes(lines[0])}"""']
    result = ['"""' + escape_escapes(lines[0])]
    result += [escape_escapes(line) for line in lines[1:]]
    result += ['"""']
    return result


def indent(lines: List[str], level: int) -> List[str]:
    """Indent each line.

    :param lines: Lines of text to be indented
    :param level: Level of indent (each indent level is four spaces)
    :return: Same lines of text provided after each line is indented
    """
    if level == 0:
        return lines
    else:
        return ['    ' * level + line if line else '' for line in lines]


def data_type(d_type: DataType) -> List[str]:
    """Generate Python code defining the provided data type.

    :param d_type: Parsed API data type to render into Python code
    :return: Lines of Python code defining the provided data type
    """
    docstring_lines = docstring_comment(
        d_type.description.split('\n')) if d_type.description else []
    lines = []

    if d_type.enum_values:
        if any(str(t) in keyword.kwlist for v, t in d_type.enum_values.items()):
            lines.append(f'{d_type.name} = {d_type.python_type}')
            docstring_lines = d_type.description.split(
                '\n') if d_type.description else []
            docstring_lines += ['', 'Acceptable values:'] + ['* ' + str(v) for v
                                                             in
                                                             d_type.enum_values]
            lines.extend(docstring_comment(docstring_lines))
        else:
            lines.append(f'class {d_type.name}({d_type.python_type}, Enum):')
            lines.extend(indent(docstring_lines, 1))
            if docstring_lines:
                lines.append('')

            lines.extend(
                indent([f'{t} = "{v}"' for v, t in d_type.enum_values.items()], 1))
    elif is_primitive_python_type(d_type.python_type):
        lines.append(f'{d_type.name} = {d_type.python_type}')
        lines.extend(docstring_lines)
    elif d_type.python_type == 'ImplicitDict':
        if any(f.api_name in keyword.kwlist for f in d_type.fields):
            lines.append(f'{d_type.name} = dict')
            if d_type.description:
                docstring_lines = d_type.description.split('\n')
                docstring_lines.append('')
            else:
                docstring_lines = []
            docstring_lines.append('Expected keys:')
            docstring_lines.extend('* ' + f.api_name for f in d_type.fields)
            lines.extend(docstring_comment(docstring_lines))
            lines.append('')
        else:
            lines.append(f'class {d_type.name}(ImplicitDict):')
            lines.extend(indent(docstring_lines, 1))
            if docstring_lines:
                lines.append('')

            for field in d_type.fields:
                lines.extend(indent(_object_field(field), 1))
                lines.append('')
            if d_type.fields:
                lines.pop()
    else:
        lines.append(f'{d_type.name} = {d_type.python_type}')

    return lines


def _object_field(field: ObjectField) -> List[str]:
    """Generate an unindented definition of the provided field in Python code.

    :param field: Data type field to render into Python code
    :return: Lines of Python code defining the provided field
    """
    d_type = field.python_type if field.required else f'Optional[{field.python_type}]'
    if field.default is not None:
        if field.literal_default:
            default_suffix = f' = {field.default}'
        elif isinstance(field.default, str):
            default_suffix = f' = "{field.default}"'
        else:
            default_suffix = f' = {str(field.default)}'
    else:
        default_suffix = ''
    lines = [f'{field.api_name}: {d_type}{default_suffix}']
    if field.description:
        lines.extend(docstring_comment(field.description.split('\n')))
    return lines


def header(d_types: List[DataType]) -> List[str]:
    lines = ['# This file is autogenerated; do not modify manually!', '']
    lines.extend(['from __future__ import annotations', ''])

    lines.append('from enum import Enum')

    basic_types = ['Dict']
    if any(d.python_type.startswith('List[') for d in d_types) or any(
            any(f.python_type.startswith('List[') for f in d.fields) for d in
            d_types):
        basic_types.append('List')
    if any(any(not f.required for f in d.fields) for d in d_types):
        basic_types.append('Optional')
    lines.append('from typing import ' + ', '.join(basic_types))
    lines.append('')

    lines.append('from uas_standards import Operation')
    lines.append('')

    lines.append('from implicitdict import ImplicitDict')
    if any(('StringBasedDateTime' in d.python_type) for d in d_types) or any(
            any(('StringBasedDateTime' in f.python_type) for f in d.fields) for
            d in d_types):
        lines[-1] = lines[-1] + ', StringBasedDateTime'

    return lines


def data_types(d_types: List[DataType], default_package: str) -> List[str]:
    already_defined = [kw for kw in keyword.kwlist]
    already_defined += ['int', 'float', 'complex', 'str', 'list', 'tuple',
                        'range', 'bytes', 'bytearray', 'memoryview', 'dict',
                        'bool', 'set', 'frozenset', 'StringBasedDateTime', 'ImplicitDict']

    lines = []

    # Declare types in dependency order
    total_defined = 0
    n_defined = 1
    not_defined = []

    def _core_type(type_name):
        core_type = type_name
        while '[' in core_type:
            core_type = core_type[core_type.index('[') + 1:]
        while ']' in core_type:
            core_type = core_type[0:core_type.index(']')]
        return core_type

    while n_defined > 0:
        n_defined = 0
        for d_type in d_types:
            if d_type.name in already_defined:
                continue
            if _core_type(d_type.python_type) not in already_defined:
                continue
            if any((_core_type(f.python_type) not in already_defined) for f in
                   d_type.fields):
                continue

            lines.extend(['', ''])
            lines.extend(data_type(d_type))
            already_defined.append(d_type.name)
            n_defined += 1
            total_defined += 1

        not_defined = [d_type for d_type in d_types if
                       d_type.name not in already_defined]
        if not not_defined:
            break

        # Declare certain types external
        if n_defined == 0:
            remaining_names = {d_type.name for d_type in d_types
                               if d_type.name not in already_defined}
            for d_type in d_types:
                only_external_undefined_fields = True
                for f in d_type.fields:
                    core_field_type = _core_type(f.python_type)
                    if core_field_type not in already_defined and core_field_type in remaining_names:
                        only_external_undefined_fields = False
                        break
                if only_external_undefined_fields:
                    for f in d_type.fields:
                        core_field_type = _core_type(f.python_type)
                        if core_field_type not in already_defined:
                            lines.extend(['', ''])
                            lines.append(f'from {default_package} import {core_field_type}')
                            already_defined.append(core_field_type)
                            n_defined += 1
                    break

    if not_defined:
        not_defined_list = '; '.join(
            [t.name + ' (' + ', '.join(_core_type(f.python_type)
                                       for f in t.fields
                                       if _core_type(f.python_type) not in already_defined) + ')'
             for t in not_defined])
        raise RuntimeError(f'Failed to define data types: {not_defined_list}')

    lines.append('')
    return lines


def operations(ops: List[Operation]) -> List[str]:
    lines = ['class OperationID(str, Enum):']
    for op in ops:
        lines.append(f'    {op.name} = "{op.id}"')
    lines.append('')
    lines.append('')

    lines.append('OPERATIONS: Dict[OperationID, Operation] = {')
    for op in ops:
        lines.append(f'    OperationID.{op.name}: Operation(')
        lines.append(f'        id="{op.id}",')
        lines.append(f'        path="{op.path}",')
        lines.append(f'        verb="{op.verb}",')
        lines.append(f'        request_body_type={op.request_body_type},')
        lines.append('        response_body_type={')
        for code, resp_body in op.response_body_type.items():
            lines.append(f'            {code}: {resp_body},')
        lines.append('        }')
        lines.append('    ),')
    lines.append('}')

    return lines



"""Data types and operations from Remote ID Display Data Observation 0.0.1 OpenAPI"""

# This file is autogenerated; do not modify manually!

from __future__ import annotations

from enum import Enum
from typing import Dict, List, Optional

from uas_standards import Operation

from implicitdict import ImplicitDict


class GetDetailsResponse(ImplicitDict):
    """Response to a request to get details about a flight."""



class Position(ImplicitDict):
    """A position on Earth."""

    lat: float
    """Degrees of latitude north of the equator, with reference to the WGS84 ellipsoid."""

    lng: float
    """Degrees of longitude east of the Prime Meridian, with reference to the WGS84 ellipsoid."""

    alt: Optional[float]
    """Geodetic altitude (NOT altitude above launch, altitude above ground, or EGM96): aircraft distance above the WGS84 ellipsoid as measured along a line that passes through the aircraft and is normal to the surface of the WGS84 ellipsoid."""


class Path(ImplicitDict):
    """Path followed by a flight."""

    positions: List[Position]
    """Sequential positions available for a flight."""


class Flight(ImplicitDict):
    id: str
    """Identifier of flight that may be used to obtain details about the flight. This is not necessarily the UTM/flight ID in the remote ID system."""

    most_recent_position: Optional[Position]
    """Most recent position known for the flight."""

    recent_paths: Optional[List[Path]]
    """Paths the flight recently traveled, if available."""


class Cluster(ImplicitDict):
    """A general area containing one or more flight."""

    corners: List[Position]
    """Two opposite corners of a rectangular lat-lng box bounding the cluster."""

    area_sqm: float
    """Area of the cluster in square meters."""

    number_of_flights: int
    """Number of flights within the cluster."""


class GetDisplayDataResponse(ImplicitDict):
    """Response to a request for current data that would be visualized by a Display Application."""

    flights: Optional[List[Flight]] = []
    """Current information for set of discovered flights whose precise locations are known."""

    clusters: Optional[List[Cluster]] = []
    """Current information for sets of discovered flights whose precise locations are not known."""


class OperationID(str, Enum):
    GetDisplayData = "getDisplayData"
    GetDetails = "getDetails"


OPERATIONS: Dict[OperationID, Operation] = {
    OperationID.GetDisplayData: Operation(
        id="getDisplayData",
        path="/display_data",
        verb="GET",
        request_body_type=None,
        response_body_type={
            200: GetDisplayDataResponse,
        }
    ),
    OperationID.GetDetails: Operation(
        id="getDetails",
        path="/display_data/{id}",
        verb="GET",
        request_body_type=None,
        response_body_type={
            200: GetDetailsResponse,
            404: None,
        }
    ),
}



# This tool generates Python data types and path constants from an OpenAPI YAML file.

import argparse

import yaml

import data_types
import operations
import rendering


def _parse_args():
    parser = argparse.ArgumentParser(description='Autogenerate Python data types from an OpenAPI YAML')

    # Input/output specifications
    parser.add_argument('--api', dest='api', type=str,
                        help='Source YAML to preprocess.')
    parser.add_argument('--python_output', dest='python_output', type=str,
                        help='Output file for generated Python code')
    parser.add_argument('--default_package', dest='default_package', type=str,
                        help='If this API refers to objects in another API, the Python package name where those other objects may be found',
                        default='<not_defined>')

    return parser.parse_args()


def main():
    args = _parse_args()

    # Parse OpenAPI
    with open(args.api, mode='r') as f:
        spec = yaml.full_load(f)

    # Parse data types
    types = data_types.parse(spec)

    # Parse operations
    ops = operations.get_operations(spec)

    # Render Python code
    with open(args.python_output, 'w') as f:
        f.write(f'"""Data types and operations from {spec["info"]["title"]} {spec["info"]["version"]} OpenAPI"""\n\n')
        f.write('\n'.join(rendering.header(types)))
        f.write('\n')
        f.write('\n'.join(rendering.data_types(types, args.default_package)))
        f.write('\n\n')
        f.write('\n'.join(rendering.operations(ops)))
        f.write('\n')


if __name__ == '__main__':
    main()



import dataclasses
from typing import Any, Dict, List, Optional, Set, Tuple


@dataclasses.dataclass
class ObjectField:
    """A data field within an Object data type"""

    api_name: str
    """Name of the field in the API (generally snake cased)"""

    python_type: str
    """The name of the Python data type which represents this field's value"""

    description: str
    """Documentation of this field"""

    required: bool
    """True if an instance of the parent object must specify a value for this field"""

    default: Optional[Any]
    """Default value for field, if specified"""

    literal_default: bool = False
    """If true, render the default without quotes even if it's a string"""


@dataclasses.dataclass
class DataType:
    """A specific data type defined in the API"""

    name: str
    """Name of this data type, as defined in the API"""

    python_type: str = ''
    """Name of the Python data type ('ImplicitDict' for Object data types)"""

    description: str = ''
    """Documentation of this data type"""

    fields: List[ObjectField] = dataclasses.field(default_factory=list)
    """If this is an Object data type, a list of fields contained in that Object"""

    enum_values: Dict[str, str] = dataclasses.field(default_factory=dict)
    """If this is a enum data type, a map from values it may take on to Python names for those values"""



python_primitives: Dict[str, str] = {
    'string': 'str',
    'boolean': 'bool',
}
"""Maps OpenAPI `type` to Python primitive type"""

python_numbers: Dict[str, str] = {
    'float': 'float',
    'double': 'float',
    'int32': 'int',
    'int64': 'int',
    'number': 'float',
    'integer': 'int',
}
"""Maps OpenAPI `format` (defaulting to `type` if `format` is missing) to Python primitive type"""


def is_primitive_python_type(python_type_name: str) -> bool:
    """True iff python_type_name describes a built-in Python type"""
    return python_type_name in python_primitives.values() or python_type_name in python_numbers.values()


def get_data_type_name(component_name: str, data_type_name: str) -> str:
    """Get the plain data type name from a $ref URI.

    :param component_name: $ref URI to the data type of interest
    :param data_type_name: context in which the data type is being retrieved (used for error message only)
    :return: Plain data type name in the relative $ref URI
    """
    if component_name == '':
        return ''
    elif component_name.startswith('#/components/schemas/'):
        return component_name[len('#/components/schemas/'):]
    else:
        if '#/components/schemas/' not in component_name:
            raise ValueError('$ref expected to contain `#/components/schemas/`, but found `{}` instead for {}'.format(component_name, data_type_name))
        name = get_data_type_name(component_name[component_name.index('#'):], data_type_name)
        print(f'WARNING: Assuming the variable type of {component_name} should be "{name}" and that it will be manually declared')
        return name


def _parse_referenced_type_name(schema: Dict, data_type_name: str) -> str:
    options = schema['anyOf'] if 'anyOf' in schema else schema['allOf']
    if len(options) != 1:
        raise NotImplementedError('Only one $ref is supported for anyOf and allOf; found {} elements instead'.format(len(options)))
    option = options[0]
    if not isinstance(option, dict):
        raise ValueError('Expected dict entries in anyOf/allOf block; found {} instead'.format(option))
    if len(option) != 1 or '$ref' not in option:
        raise NotImplementedError('The only element in anyOf/allOf must be a $ref dictionary; found {} instead'.format(option))
    return get_data_type_name(option['$ref'], data_type_name)


def make_object_field(python_object_name: str, api_field_name: str, schema: Dict, required: Set[str]) -> Tuple[ObjectField, List[DataType]]:
    """Parse a single field in a data type or endpoint parameter schema.

    :param python_object_name: Name of the Python object containing this field, for error messages and inline type names
    :param api_field_name: Name of the object field being parsed, according to the API
    :param schema: Definition of the object field being parsed
    :param required: The set of required fields for the parent object
    :return: Tuple of
      * The object field defined by the provided schema
      * Any additional data types incidentally defined in the provided schema
    """
    is_required = api_field_name in required
    default_value = schema['default'] if 'default' in schema else None
    if '$ref' in schema:
        return ObjectField(
            api_name=api_field_name,
            python_type=get_data_type_name(schema['$ref'], python_object_name),
            description=schema.get('description', ''),
            required=is_required,
            default=default_value), []
    elif 'anyOf' in schema or 'allOf' in schema:
        return ObjectField(
            api_name=api_field_name,
            python_type=_parse_referenced_type_name(schema, python_object_name + '.' + api_field_name),
            description=schema.get('description', ''),
            required=is_required,
            default=default_value), []
    else:
        type_name = python_object_name + api_field_name[0].upper() + api_field_name[1:]
        data_type, additional_types = make_data_types(type_name, schema)
        if is_primitive_python_type(data_type.python_type) and not data_type.enum_values:
            # No additional type declaration needed
            if additional_types:
                raise RuntimeError('{} field type `{}` was parsed as primitive {} but also generated {} additional types'.format(python_object_name, api_field_name, data_type.python_type, len(additional_types)))
            field_data_type = data_type.python_type
        elif data_type.python_type == 'StringBasedDateTime':
            # No additional type declaration needed
            field_data_type = data_type.python_type
        elif data_type.python_type.startswith('List['):
            # Use array data type as-is
            field_data_type = data_type.python_type
        else:
            additional_types.append(data_type)
            field_data_type = data_type.name
        if len(data_type.enum_values) == 1 and default_value is None:
            default_value = data_type.name + '.' + next(iter(data_type.enum_values.values()))
            literal_default = True
        else:
            literal_default = False
        return ObjectField(
            api_name=api_field_name,
            python_type=field_data_type,
            description=data_type.description,
            required=is_required,
            default=default_value,
            literal_default=literal_default), additional_types


def _make_object_fields(python_object_name: str, properties: Dict, required: Set[str]) -> Tuple[List[ObjectField], List[DataType]]:
    fields: List[ObjectField] = []
    additional_types: List[DataType] = []
    for field_name, schema in properties.items():
        field, further_types = make_object_field(python_object_name, field_name, schema, required)
        additional_types.extend(further_types)
        fields.append(field)
    return fields, additional_types


def _make_python_enums(values: List[str]) -> Dict[str, str]:
    valid_characters = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_'
    valid_start_characters = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_'
    result = {}
    for value in values:
        name = str(value)
        if name[0] not in valid_start_characters:
            name = '_' + name
        name = ''.join(c if c in valid_characters else '_' for c in name)
        result[str(value)] = name
    return result


def make_data_types(api_name: str, schema: Dict) -> Tuple[DataType, List[DataType]]:
    """Parse all data types necessary to express the provided data type schema.

    In addition to the primary data type described by `name`, this routine also
    generates additional data types defined inline in the provided schema.

    :param api_name: Name of the primary data type being parsed, according to the API
    :param schema: Definition of the data type being parsed
    :return: Tuple of
      * The primary data defined by the provided schema
      * Any additional data types incidentally defined in the provided schema
    """
    data_type = DataType(name=api_name)
    additional_types = []

    if 'description' in schema:
        data_type.description = schema['description']

    if 'type' in schema:
        if schema['type'] in python_primitives:
            data_type.python_type = python_primitives[schema['type']]
            if data_type.python_type == 'str' and 'format' in schema and schema['format'] == 'date-time':
                data_type.python_type = 'StringBasedDateTime'
        elif schema['type'] in {'number', 'integer'}:
            data_type.python_type = python_numbers.get(schema.get('format', schema['type']), '')
            if not data_type.python_type:
                raise ValueError('Unrecognized numeric format `{}` for {}'.format(schema.get('format', '<missing>'), api_name))
        elif schema['type'] == 'array':
            if 'items' in schema:
                items = schema['items']
                if '$ref' in items:
                    item_type_name = get_data_type_name(items['$ref'], api_name)
                else:
                    item_type, further_types = make_data_types(api_name + 'Item', items)
                    additional_types.extend(further_types)
                    if item_type.description != '' or not is_primitive_python_type(item_type.python_type):
                        additional_types.append(item_type)
                        item_type_name = item_type.name
                    else:
                        item_type_name = item_type.python_type
                data_type.python_type = f'List[{item_type_name}]'
            else:
                raise ValueError('Missing `items` declaration for {} array type'.format(api_name))
        elif schema['type'] == 'object':
            data_type.python_type = 'ImplicitDict'
            data_type.fields, further_types = _make_object_fields(
                api_name,
                schema.get('properties', {}),
                set(schema.get('required', [])))
            additional_types.extend(further_types)
        else:
            raise ValueError('Unrecognized type `{}` in {} type'.format(schema['type'], api_name))
    elif 'anyOf' in schema or 'allOf' in schema:
        data_type.python_type = _parse_referenced_type_name(schema, api_name)

    if 'enum' in schema:
        data_type.enum_values = _make_python_enums(schema['enum'])

    if not data_type.python_type:
        data_type.python_type = 'str'

    return data_type, additional_types


def parse(spec: dict) -> List[DataType]:
    if 'components' not in spec:
        raise ValueError('Missing `components` in YAML')
    components = spec['components']
    if 'schemas' not in components:
        raise ValueError('Missing `schemas` in `components`')
    declared_types = []
    for name, schema in components['schemas'].items():
        data_type, additional_types = make_data_types(name, schema)
        declared_types.extend(additional_types)
        declared_types.append(data_type)
    return declared_types



from __future__ import annotations
import random
from typing import Optional


class SerialNumber(str):
    """Represents a serial number expressed in the ANSI/CTA-2063-A Physical Serial Number format."""

    length_code_points = "123456789ABCDEF"
    code_points = "0123456789ABCDEFGHJKLMNPQRSTUVWXYZ"

    @property
    def manufacturer_code(self) -> str:
        return self[0:4]

    @property
    def length_code(self) -> str:
        return self[4:5]

    @property
    def manufacturer_serial_number(self) -> str:
        return self[5:]

    @property
    def valid(self) -> bool:
        if len(self) < 6:
            return False
        if not all(c in SerialNumber.code_points for c in self.manufacturer_code):
            return False
        if self.length_code not in SerialNumber.length_code_points:
            return False
        manufacturer_serial_number_length = (
            SerialNumber.length_code_points.index(self.length_code) + 1
        )
        if manufacturer_serial_number_length != len(self.manufacturer_serial_number):
            return False
        return True

    def make_invalid_by_changing_payload_length(self, r: Optional[random.Random] = None) -> SerialNumber:
        """Generates an invalid serial number similar to this serial number."""
        if r is None:
            r = random
        my_length = self.length_code
        lengths_except_mine = [
            c for c in SerialNumber.length_code_points if c != my_length
        ]
        new_length_code = r.choice(lengths_except_mine)
        k = SerialNumber.length_code_points.index(new_length_code) + 1
        while True:
            random_serial_number = "".join(r.choices(SerialNumber.code_points, k=k))
            result = SerialNumber(
                self.manufacturer_code + self.length_code + random_serial_number
            )
            if not result.valid:
                return result

    @staticmethod
    def from_components(
        manufacturer_code: str, manufacturer_serial_number: str
    ) -> SerialNumber:
        """Constructs a standard serial number from the provided components"""
        length_code = SerialNumber.length_code_points[
            len(manufacturer_serial_number) - 1
        ]
        return SerialNumber(
            manufacturer_code + length_code + manufacturer_serial_number
        )

    @staticmethod
    def generate_valid(r: Optional[random.Random] = None) -> SerialNumber:
        """Generates a valid and random UAV serial number per ANSI/CTA-2063-A."""
        if r is None:
            r = random
        manufacturer_code = "".join(r.choices(SerialNumber.code_points, k=4))
        k = r.randrange(0, len(SerialNumber.length_code_points)) + 1
        random_serial_number = "".join(r.choices(SerialNumber.code_points, k=k))
        return SerialNumber.from_components(manufacturer_code, random_serial_number)



"""Data types and operations from Standard Remote ID API Interfaces 2.1.0 OpenAPI"""

# This file is autogenerated; do not modify manually!

from __future__ import annotations

from enum import Enum
from typing import Dict, List, Optional

from uas_standards import Operation

from implicitdict import ImplicitDict, StringBasedDateTime


class TimeFormat(str, Enum):
    RFC3339 = "RFC3339"


class Time(ImplicitDict):
    value: StringBasedDateTime
    """RFC3339-formatted time/date string.  The time zone must be 'Z'."""

    format: TimeFormat = TimeFormat.RFC3339


class RadiusUnits(str, Enum):
    """FIXM-compatible units.  Only meters ("M") are acceptable for UTM."""

    M = "M"


class Radius(ImplicitDict):
    value: float
    """Distance from the centerpoint of a circular area, along the WGS84 ellipsoid."""

    units: RadiusUnits = RadiusUnits.M
    """FIXM-compatible units.  Only meters ("M") are acceptable for UTM."""


URL = str
"""Valid http or https URL."""


SubscriptionNotificationIndex = int
"""Tracks the notifications sent for a subscription so the subscriber can detect missed notifications more easily."""


UUIDv4 = str
"""UUID v4."""


Version = str
"""A version string used to reference an object at a particular point in time. Any updates to an object must contain the corresponding version to maintain idempotent updates."""


EntityUUID = UUIDv4


SubscriptionUUID = UUIDv4


RIDFlightID = str
"""ID, unique to a remote ID service provider, which identifies a particular flight for which the remote ID service provider is providing remote ID services.

The following characters are not allowed: \\0 \\t \\r \\n # % / : ? @ [ \\ ]
"""


class RIDAuthData(ImplicitDict):
    """Additional authentication data."""

    format: Optional[int] = 0
    """Format of additional authentication data.

    0: None
    1: UAS ID Signature
    2: Operator ID Signature
    3: Message Set Signature
    4: Authentication Provided by Network Remote ID
    5: Specific Method
    6-9: Reserved for Spec
    10-15: Available for Private Use
    """

    data: Optional[str] = ""
    """Authentication data in form specified by `format`."""


class HorizontalAccuracy(str, Enum):
    """This is based on ADS-B NACp (plus the one extra increment).

    `HAUnknown`: Unknown horizontal accuracy

    `HA10NMPlus`: > 10NM (18.52km)

    `HA10NM`: < 10NM (18.52km)

    `HA4NM`: < 4NM (7.408km)

    `HA2NM`: < 2NM (3.704km)

    `HA1NM`: < 1NM (1852m)

    `HA05NM`: < 0.5NM (926m)

    `HA03NM`: < 0.3NM (555.6m)

    `HA01NM`: < 0.1NM (185.2m)

    `HA005NM`: < 0.05NM (92.6m)

    `HA30m`: < 30m

    `HA10m`: < 10m

    `HA3m`: < 3m

    `HA1m`: < 1m
    """

    HAUnknown = "HAUnknown"
    HA10NMPlus = "HA10NMPlus"
    HA10NM = "HA10NM"
    HA4NM = "HA4NM"
    HA2NM = "HA2NM"
    HA1NM = "HA1NM"
    HA05NM = "HA05NM"
    HA03NM = "HA03NM"
    HA01NM = "HA01NM"
    HA005NM = "HA005NM"
    HA30m = "HA30m"
    HA10m = "HA10m"
    HA3m = "HA3m"
    HA1m = "HA1m"


class VerticalAccuracy(str, Enum):
    """This is based on ADS-B Geodetic Vertical Accuracy (GVA) (plus the three extra increments)

    `VAUnknown`: Unknown vertical accuracy

    `VA150mPlus`: > 150m

    `VA150m`: < 150m

    `VA45m`: < 45m

    `VA25m`: < 25m

    `VA10m`: < 10m

    `VA3m`: < 3m

    `VA1m`: < 1m
    """

    VAUnknown = "VAUnknown"
    VA150mPlus = "VA150mPlus"
    VA150m = "VA150m"
    VA45m = "VA45m"
    VA25m = "VA25m"
    VA10m = "VA10m"
    VA3m = "VA3m"
    VA1m = "VA1m"


class ErrorResponse(ImplicitDict):
    """Data provided when an off-nominal condition was encountered."""

    message: Optional[str] = ""
    """Human-readable message indicating what error occurred and/or why."""


class SpeedAccuracy(str, Enum):
    """Provides quality/containment on horizontal ground speed.

    `SAUnknown`: Unknown speed accuracy

    `SA10mpsPlus`: > 10 m/s

    `SA10mps`: < 10 m/s

    `SA3mps`: < 3 m/s

    `SA1mps`: < 1 m/s

    `SA03mps`: < 0.3 m/s
    """

    SAUnknown = "SAUnknown"
    SA10mpsPlus = "SA10mpsPlus"
    SA10mps = "SA10mps"
    SA3mps = "SA3mps"
    SA1mps = "SA1mps"
    SA03mps = "SA03mps"


GeoPolygonString = str
"""Plain-string representation of a geographic polygon consisting of at least three geographic
points describing a closed polygon on the earth.  Each point consists of latitude,longitude
in degrees.  Points are also comma-delimited, so this parameter will look like
`lat1,lng1,lat2,lng2,lat3,lng3,...`  Latitude values must fall in the range [-90, 90] and
longitude values must fall in the range [-180, 180].

All of the requirements and clarifications for Polygon apply to GeoPolygonString as well.
"""


class RIDHeightReference(str, Enum):
    """The reference datum above which the height is reported."""

    TakeoffLocation = "TakeoffLocation"
    GroundLevel = "GroundLevel"


class RIDHeight(ImplicitDict):
    """A relative altitude for the purposes of remote ID."""

    distance: Optional[float] = 0
    """Distance above reference datum.  This value is provided in meters and must have a minimum resolution of 1 meter.  Invalid, No Value or Unknown is -1000 m."""

    reference: RIDHeightReference
    """The reference datum above which the height is reported."""


Latitude = float
"""Degrees of latitude north of the equator, with reference to the WGS84 ellipsoid.  Invalid, No Value, or Unknown is 0 degrees (both Lat/Lon)."""


Longitude = float
"""Degrees of longitude east of the Prime Meridian, with reference to the WGS84 ellipsoid.  Invalid, No Value, or Unknown is 0 degrees (both Lat/Lon)."""


class LatLngPoint(ImplicitDict):
    """Point on the earth's surface."""

    lng: Longitude

    lat: Latitude


class AltitudeReference(str, Enum):
    """A code indicating the reference for a vertical distance. See AIXM 5.1 and FIXM 4.2.0. Currently, UTM only allows WGS84 with no immediate plans to allow other options. FIXM and AIXM allow for 'SFC' which is equivalent to AGL."""

    W84 = "W84"


class AltitudeUnits(str, Enum):
    """The reference quantities used to express the value of altitude. See FIXM 4.2. Currently, UTM only allows meters with no immediate plans to allow other options."""

    M = "M"


class Altitude(ImplicitDict):
    value: float
    """The numeric value of the altitude. Note that min and max values are added as a sanity check. As use cases evolve and more options are made available in terms of units of measure or reference systems, these bounds may be re-evaluated. Invalid, No Value, or Unknown is –1000 m."""

    reference: AltitudeReference = AltitudeReference.W84
    """A code indicating the reference for a vertical distance. See AIXM 5.1 and FIXM 4.2.0. Currently, UTM only allows WGS84 with no immediate plans to allow other options. FIXM and AIXM allow for 'SFC' which is equivalent to AGL."""

    units: AltitudeUnits = AltitudeUnits.M
    """The reference quantities used to express the value of altitude. See FIXM 4.2. Currently, UTM only allows meters with no immediate plans to allow other options."""


class Polygon(ImplicitDict):
    """An enclosed area on the earth. The bounding edges of this polygon are defined to be the shortest paths between connected vertices.  This means, for instance, that the edge between two points both defined at a particular latitude is not generally contained at that latitude. The winding order must be interpreted as the order which produces the smaller area. The path between two vertices is defined to be the shortest possible path between those vertices. Edges may not cross. Vertices may not be duplicated.  In particular, the final polygon vertex must not be identical to the first vertex."""

    vertices: List[LatLngPoint]


class RIDOperationalStatus(str, Enum):
    """Operational Status indicates whether the associated UA is on the ground, airborne, or in an
    emergency situation, or the Remote ID system has failed. The emergency status takes
    precedence over the other status modes. This status can be used for filtering purposes.
    """

    Undeclared = "Undeclared"
    Ground = "Ground"
    Airborne = "Airborne"
    Emergency = "Emergency"
    RemoteIDSystemFailure = "RemoteIDSystemFailure"


SpecificSessionID = str
"""A unique 20 byte ID intended to identify a specific flight (session) while providing a
greater level of privacy to the operator. The first byte is the registered unique Specific Session ID
Type maintained by a registrar (see Annex A5), and the remaining 19 bytes is the Session ID per the
Specific Session ID Type specification.

Initial scheme registry entries include:

0 – reserved
1 – Internet Engineering Task Force (IETF) Drone Remote Identification Protocol (DRIP) entity ID
2 – IEEE 1609.2 HashedID8

Expressed as a hexadecimal string of the underlying data bytes.  A recipient should ignore any dashes in this value.  If fewer than 20 bytes are specified, the remaining bytes at the end are assumed to have value 0.
"""


class UASID(ImplicitDict):
    """Identification of the UAS performing this flight.  At least one field of this object must be specified."""

    serial_number: Optional[str] = ""
    """This is expressed in the CTA-2063-A Serial Number format."""

    registration_id: Optional[str] = ""
    """If a CAA provides a method of registering UAS, this number is provided by the CAA or its authorized representative.  Format is <ICAO Nationality MarkA>.<CAA Assigned ID>, ASCII encoded, only uppercase letters (A-Z), dot (.), and digits (0-9) are allowed. Example is valid for the US."""

    utm_id: Optional[str] = ""
    """A UTM-provided universally unique ID traceable to a non-obfuscated ID that acts as a "session id" to protect exposure of operationally sensitive information."""

    specific_session_id: Optional[SpecificSessionID]


class OperatorLocationAltitude_type(str, Enum):
    """Source of data for the altitude field."""

    Takeoff = "Takeoff"
    Dynamic = "Dynamic"
    Fixed = "Fixed"


class OperatorLocation(ImplicitDict):
    """Location associated with the Remote Pilot"""

    position: LatLngPoint
    """Position of the Remote Pilot."""

    altitude: Optional[Altitude]
    """Provides the Operator Altitude based on WGS-84 height above ellipsoid (HAE) (See Geodetic Altitude).  This value is provided in meters and must have a minimum resolution of 1 m."""

    altitude_type: Optional[OperatorLocationAltitude_type]
    """Source of data for the altitude field."""


class UAClassificationEUCategory(str, Enum):
    EUCategoryUndefined = "EUCategoryUndefined"
    Open = "Open"
    Specific = "Specific"
    Certified = "Certified"


class UAClassificationEUClass(str, Enum):
    EUClassUndefined = "EUClassUndefined"
    Class0 = "Class0"
    Class1 = "Class1"
    Class2 = "Class2"
    Class3 = "Class3"
    Class4 = "Class4"
    Class5 = "Class5"
    Class6 = "Class6"


UAClassificationEU = dict
"""Expected keys:
* category
* class
"""



class RIDFlightDetails(ImplicitDict):
    """Details about a flight reported by a remote ID service provider.  At least one of the registration or serial fields must be filled if required by CAA."""

    id: str
    """ID for this flight, matching argument in request."""

    eu_classification: Optional[UAClassificationEU]
    """When this field is specified, the Classification Type is "European Union".  If no other classification field is specified, the Classification Type is "Undeclared"."""

    uas_id: Optional[UASID]

    operator_id: Optional[str] = ""
    """This optional field provides a CAA-issued registration/license ID for the remote pilot or operator. Format is <ICAO Nationality MarkA>.<CAA Assigned ID>, ASCII encoded, only uppercase letters (A-Z), dot (.), and digits (0-9) are allowed. Example is valid for the US."""

    operator_location: Optional[OperatorLocation]

    operation_description: Optional[str] = ""
    """Free-text field that enables the operator to describe the purpose of a flight, if so desired."""

    auth_data: Optional[RIDAuthData]


class UAType(str, Enum):
    """The UA Type can help infer performance, speed, and duration of flights, for example, a
    "fixed wing" can generally fly in a forward direction only (as compared to a multi-rotor).
    This can also help differentiate aircraft types without sharing operationally sensitive
    information like the make and model of a particular aircraft. Make and model are
    anticipated to become available during the Registration ID lookup process. UAS Type is also
    useful for correlating visual observation with data received. The types were formulated
    based on unique flight characteristics.

    `HybridLift` is a fixed wing aircraft that can take off vertically.  `Helicopter` includes multirotor.
    """

    NotDeclared = "NotDeclared"
    Aeroplane = "Aeroplane"
    Helicopter = "Helicopter"
    Gyroplane = "Gyroplane"
    HybridLift = "HybridLift"
    Ornithopter = "Ornithopter"
    Glider = "Glider"
    Kite = "Kite"
    FreeBalloon = "FreeBalloon"
    CaptiveBalloon = "CaptiveBalloon"
    Airship = "Airship"
    FreeFallOrParachute = "FreeFallOrParachute"
    Rocket = "Rocket"
    TetheredPoweredAircraft = "TetheredPoweredAircraft"
    GroundObstacle = "GroundObstacle"
    Other = "Other"


USSBaseURL = str
"""The base URL of a USS implementation of part or all of the remote ID USS-USS API. Per the USS-USS API, the full URL
of a particular resource can be constructed by appending, e.g., `/uss/{resource}/{id}` to this base URL.
Accordingly, this URL may not have a trailing '/' character.
"""


SubscriptionUSSBaseURL = USSBaseURL


FlightsUSSBaseURL = USSBaseURL


class Circle(ImplicitDict):
    """A circular area on the surface of the earth."""

    center: Optional[LatLngPoint]

    radius: Optional[Radius]


class Volume3D(ImplicitDict):
    """A three-dimensional geographic volume consisting of a vertically-extruded shape. Exactly one outline must be specified."""

    outline_circle: Optional[Circle]
    """A circular geographic shape on the surface of the earth."""

    outline_polygon: Optional[Polygon]
    """A polygonal geographic shape on the surface of the earth."""

    altitude_lower: Optional[Altitude]
    """Minimum bounding altitude of this volume. Must be less than altitude_upper, if specified."""

    altitude_upper: Optional[Altitude]
    """Maximum bounding altitude of this volume. Must be greater than altitude_lower, if specified."""


class Volume4D(ImplicitDict):
    """Contiguous block of geographic spacetime."""

    volume: Volume3D

    time_start: Optional[Time]
    """Beginning time of this volume. Must be before time_end."""

    time_end: Optional[Time]
    """End time of this volume. Must be after time_start."""


class SubscriptionState(ImplicitDict):
    """State of Subscription which is causing a notification to be sent."""

    subscription_id: SubscriptionUUID

    notification_index: Optional[SubscriptionNotificationIndex] = 0


class GetFlightDetailsResponse(ImplicitDict):
    """Response to remote ID provider query for details about a specific flight."""

    details: RIDFlightDetails


class RIDAircraftPosition(ImplicitDict):
    """Position of an aircraft as reported for remote ID purposes."""

    lat: Optional[Latitude]

    lng: Optional[Longitude]

    alt: Optional[float] = -1000
    """Geodetic altitude (NOT altitude above launch, altitude above ground, or EGM96): aircraft distance above the WGS84 ellipsoid as measured along a line that passes through the aircraft and is normal to the surface of the WGS84 ellipsoid.  This value is provided in meters and must have a minimum resolution of 1 meter.  Invalid, No Value or Unknown is -1000 m."""

    accuracy_h: Optional[HorizontalAccuracy]
    """Horizontal error that is likely to be present in this reported position.  Required when `extrapolated` field is true and always in the entry for the current state."""

    accuracy_v: Optional[VerticalAccuracy]
    """Vertical error that is likely to be present in this reported position.  Required when `extrapolated` field is true and always in the entry for the current state."""

    extrapolated: Optional[bool] = False
    """True if this position was generated primarily by computation rather than primarily from a direct instrument measurement.  Assumed false if not specified."""

    pressure_altitude: Optional[float] = -1000
    """The uncorrected altitude (based on reference standard 29.92 inHg, 1013.25 mb) provides a reference for algorithms that utilize "altitude deltas" between aircraft.  This value is provided in meters and must have a minimum resolution of 1 meter.  Invalid, No Value or Unknown is -1000 m."""

    height: Optional[RIDHeight]


class OperatingArea(ImplicitDict):
    """Area of operation containing one or more aircraft participating in remote identification."""

    aircraft_count: Optional[int]
    """Allows for operating a single UA, group, formation, or swarm. Quantity in Group."""

    volumes: Optional[List[Volume4D]] = []
    """The area where a group or Intent-Based Network Participant operation is planned or taking place."""


class SubscriberToNotify(ImplicitDict):
    """Subscriber to notify of a creation/change/deletion of a change in the airspace.  This is provided by the DSS to a client changing the airspace, and it is the responsibility of the client changing the airspace (they will receive a set of these notification requests) to send a notification to each specified `url`."""

    subscriptions: List[SubscriptionState]
    """Subscription(s) prompting this notification."""

    url: URL
    """The endpoint that the client mutating the airspace should provide the update to.  API depends on the DSS action taken that triggered this notification request."""


class RIDRecentAircraftPosition(ImplicitDict):
    time: Time
    """Time at which this position applied."""

    position: RIDAircraftPosition


class GetIdentificationServiceAreaDetailsResponse(ImplicitDict):
    """Response to request for the details of an identification service area with the given ID."""

    extents: Volume4D
    """The extents of the Identification Service Area."""


class CreateIdentificationServiceAreaParameters(ImplicitDict):
    """Parameters for a request to create an Identification Service Area in the DSS."""

    extents: Volume4D
    """The bounding spacetime extents of this Identification Service Area.  End time must be specified.  If start time is not specified, it will be set to the current time.  Start times in the past should be rejected by the DSS, except that it may adjust very recent start times to the current time.

    These extents should not reveal any sensitive information about the flight or flights within them.  This means, for instance, that extents should not tightly-wrap a flight path, nor should they generally be centered around the takeoff point of a single flight.
    """

    uss_base_url: FlightsUSSBaseURL


class UpdateIdentificationServiceAreaParameters(ImplicitDict):
    """Parameters for a request to update an Identification Service Area in the DSS."""

    extents: Volume4D
    """The bounding spacetime extents of this Identification Service Area.  End time must be specified.  If start time is not specified, it will remain unchanged.  Start times in the past should be rejected by the DSS unless they are unchanged from the Identification Service Area's current start time.

    These extents should not reveal any sensitive information about the flight or flights within them.  This means, for instance, that extents should not tightly-wrap a flight path, nor should they generally be centered around the takeoff point of a single flight.
    """

    uss_base_url: FlightsUSSBaseURL


class CreateSubscriptionParameters(ImplicitDict):
    """Parameters for a request to create a subscription in the DSS."""

    extents: Volume4D
    """The spacetime extents of the volume to subscribe to.

    This subscription will automatically be deleted after its end time if it has not been refreshed by then.  If end time is not specified, the value will be chosen automatically by the DSS.

    Note that some Entities triggering notifications may lie entirely outside the requested area.
    """

    uss_base_url: SubscriptionUSSBaseURL


class UpdateSubscriptionParameters(ImplicitDict):
    """Parameters for a request to update a subscription in the DSS."""

    extents: Volume4D
    """The spacetime extents of the volume to subscribe to.

    This subscription will automatically be deleted after its end time if it has not been refreshed by then.  If end time is not specified, the value will be chosen automatically by the DSS.

    Note that some Entities triggering notifications may lie entirely outside the requested area.
    """

    uss_base_url: SubscriptionUSSBaseURL


class Subscription(ImplicitDict):
    """Specification of a geographic area that a client is interested in on an ongoing basis (e.g., "planning area").  Internal to the DSS."""

    id: SubscriptionUUID
    """Unique identifier for this subscription."""

    uss_base_url: SubscriptionUSSBaseURL

    owner: str
    """Assigned by the DSS based on creating client’s ID (via access token).  Used for restricting mutation and deletion operations to owner."""

    notification_index: Optional[SubscriptionNotificationIndex] = 0

    time_end: Optional[Time]
    """If set, this subscription will be automatically removed after this time."""

    time_start: Optional[Time]
    """If set, this Subscription will not generate any notifications before this time."""

    version: Version


class IdentificationServiceArea(ImplicitDict):
    """An Identification Service Area (area in which remote ID services are being provided).  The DSS reports only these declarations and clients must exchange flight information peer-to-peer."""

    uss_base_url: FlightsUSSBaseURL

    owner: str
    """Assigned by the DSS based on creating client’s ID (via access token).  Used for restricting mutation and deletion operations to owner."""

    time_start: Time
    """Beginning time of service."""

    time_end: Time
    """End time of service."""

    version: Version

    id: EntityUUID
    """Unique identifier for this Identification Service Area."""


class RIDAircraftState(ImplicitDict):
    """State of an aircraft for the purposes of remote ID."""

    timestamp: Time
    """Time at which this state was valid.  This may be the time coming from the source, such as a GPS, or the time when the system computes the values using an algorithm such as an Extended Kalman Filter (EKF).  Timestamp must be expressed with a minimum resolution of 1/10th of a second."""

    timestamp_accuracy: float = 0
    """Declaration of timestamp accuracy, which is the largest difference between Timestamp and true time of applicability for any of the following fields: Latitude, Longitude, Geodetic Altitude, Pressure Altitude of Position, Height. to determine time of applicability of the location data provided.  Expressed in seconds, precise to 1/10ths of seconds. The accuracy reflects the 95% uncertainty bound value for the timestamp."""

    operational_status: Optional[RIDOperationalStatus]

    position: RIDAircraftPosition

    track: Optional[float] = 361
    """Direction of flight expressed as a "True North-based" ground track angle.  This value is provided in clockwise degrees with a minimum resolution of 1 degree.  If aircraft is not moving horizontally, use the "Unknown" value.  A value of 361 indicates invalid, no value, or unknown."""

    speed: Optional[float] = 255
    """Ground speed of flight in meters per second.  Invalid, No Value, or Unknown is 255 m/s, if speed is >254.25 m/s then report 254.25 m/s."""

    speed_accuracy: SpeedAccuracy
    """Provides quality/containment on horizontal ground speed."""

    vertical_speed: Optional[float] = 63
    """Speed up (vertically) WGS84-HAE, m/s.  Invalid, No Value, or Unknown is 63 m/s, if speed is >62 m/s then report 62 m/s."""


class GetSubscriptionResponse(ImplicitDict):
    """Response to DSS request for the subscription with the given id."""

    subscription: Subscription


class SearchSubscriptionsResponse(ImplicitDict):
    """Response to DSS query for subscriptions in a particular area."""

    subscriptions: Optional[List[Subscription]] = []
    """Subscriptions that overlap the specified area."""


class DeleteSubscriptionResponse(ImplicitDict):
    """Response for a successful request to delete an Subscription."""

    subscription: Subscription
    """The Subscription which was deleted."""


class RIDFlight(ImplicitDict):
    """Description of a remote ID flight."""

    id: RIDFlightID

    aircraft_type: UAType
    """Generic type of aircraft."""

    current_state: Optional[RIDAircraftState]
    """The most up-to-date state of the aircraft.  Required when the aircraft is
    currently in the requested area unless no telemetry is expected during during
    the flight and `volumes` is specified.

    If current data is not being received from the UAS by the Service Provider,
    the lack of change in this field is sufficient to indicate that current
    data is not being received.

    If the USS believes a flight is airborne and expects to receive telemetry
    at some point during the flight but has not yet received telemetry, this
    field should be populated with the USS's best estimate of the position
    of the flight with appropriate accuracy indications.
    """

    operating_area: Optional[OperatingArea]
    """The area the aircraft is/are within.  Required, with 1 or more elements, if
    `current_state` is not specified.  The fields `time_start` and `time_end` are
    required.

    If `current_state` is specified, this field should not be specified.
    If `current_state` was available at any time during the flight, or is
    expected to be available at any time during the flight, this field should
    not be specified.
    """

    simulated: Optional[bool] = False
    """If specified as true, this flight is not a physical aircraft; it is just a simulation to test the system."""

    recent_positions: Optional[List[RIDRecentAircraftPosition]] = []
    """A short collection of recent aircraft movement, specified only when `recent_positions_duration` is greater than zero.  If `volumes` is not specified and `recent_positions_duration` is greater than zero, then this field is required.

    Recent positions provided in this field must conform to requirements in the standard which generally prohibit including positions outside the requested area except transitionally when the aircraft enters or exits the requested area, and which prohibit including positions that not sufficiently recent.

    Note that a UI should not draw a connective line between two consecutive position reports that both lie outside the requested area.
    """


class PutIdentificationServiceAreaResponse(ImplicitDict):
    """Response to a request to create or update a reference to an Identification Service Area in the DSS."""

    subscribers: Optional[List[SubscriberToNotify]] = []
    """DSS subscribers that this client now has the obligation to notify of the Identification Service Area changes just made.  This client must call POST for each provided URL according to the `/uss/identification_service_areas/{id}` path API."""

    service_area: IdentificationServiceArea
    """Resulting service area stored in DSS."""


class SearchIdentificationServiceAreasResponse(ImplicitDict):
    """Response to DSS query for Identification Service Areas in an area of interest."""

    service_areas: Optional[List[IdentificationServiceArea]] = []
    """Identification Service Areas in the area of interest."""


class PutIdentificationServiceAreaNotificationParameters(ImplicitDict):
    """Parameters of a message informing of new full information for an Identification Service Area.  Pushed (by a client, not the DSS) directly to clients with subscriptions when another client makes a change to airspace within a cell with a subscription."""

    service_area: Optional[IdentificationServiceArea]
    """Identification Service Area that the notifying client changed or created.

    If this field is populated, the Identification Service Area was created or updated.  If this field is not populated, the Identification Service Area was deleted.
    """

    subscriptions: List[SubscriptionState]
    """Subscription(s) prompting this notification."""

    extents: Optional[Volume4D]
    """The new or updated extents of the Identification Service Area.

    Omitted if Identification Service Area was deleted.
    """


class DeleteIdentificationServiceAreaResponse(ImplicitDict):
    """Response for a request to delete an Identification Service Area."""

    service_area: IdentificationServiceArea
    """Identification Service Area that was just deleted."""

    subscribers: Optional[List[SubscriberToNotify]] = []
    """DSS subscribers that this client now has the obligation to notify of the Identification Service Area just deleted.  This client must call POST for each provided URL according to the `/uss/identification_service_areas` path API."""


class PutSubscriptionResponse(ImplicitDict):
    """Response for a request to create or update a subscription."""

    service_areas: Optional[List[IdentificationServiceArea]] = []
    """Identification Service Areas in or near the subscription area at the time of creation/update, if `identification_service_area_url` callback was specified."""

    subscription: Subscription
    """Result of the operation on the subscription."""


class GetIdentificationServiceAreaResponse(ImplicitDict):
    """Response to DSS request for the identification service area with the given ID."""

    service_area: IdentificationServiceArea


class GetFlightsResponse(ImplicitDict):
    """Response to remote ID provider query for flight information in an area of interest."""

    timestamp: Time
    """The remote ID service provider's timestamp for when this information was current."""

    flights: Optional[List[RIDFlight]] = []
    """A list of all flights that have been within the requested area within the remote ID retention period.  This includes flights that are not currently within the requested area, but were within the requested area within the remote ID retention period.  Aircraft that are not in flight, and have not been in flight within the retention period, should not be included in this list."""

    no_isas_present: Optional[bool] = False
    """The requested view is entirely outside any remote ID service provision areas (Identification Service Area volumes) this provider has.  A display provider receiving this response should discontinue polling for this endpoint for the view port requested until otherwise directed by DSS interactions."""


class OperationID(str, Enum):
    SearchIdentificationServiceAreas = "searchIdentificationServiceAreas"
    GetIdentificationServiceArea = "getIdentificationServiceArea"
    CreateIdentificationServiceArea = "createIdentificationServiceArea"
    UpdateIdentificationServiceArea = "updateIdentificationServiceArea"
    DeleteIdentificationServiceArea = "deleteIdentificationServiceArea"
    SearchFlights = "searchFlights"
    GetFlightDetails = "getFlightDetails"
    SearchSubscriptions = "searchSubscriptions"
    GetSubscription = "getSubscription"
    CreateSubscription = "createSubscription"
    UpdateSubscription = "updateSubscription"
    DeleteSubscription = "deleteSubscription"
    GetIdentificationServiceAreaDetails = "getIdentificationServiceAreaDetails"
    PostIdentificationServiceArea = "postIdentificationServiceArea"


OPERATIONS: Dict[OperationID, Operation] = {
    OperationID.SearchIdentificationServiceAreas: Operation(
        id="searchIdentificationServiceAreas",
        path="/dss/identification_service_areas",
        verb="GET",
        request_body_type=None,
        response_body_type={
            200: SearchIdentificationServiceAreasResponse,
            400: ErrorResponse,
            401: ErrorResponse,
            403: ErrorResponse,
            413: ErrorResponse,
        }
    ),
    OperationID.GetIdentificationServiceArea: Operation(
        id="getIdentificationServiceArea",
        path="/dss/identification_service_areas/{id}",
        verb="GET",
        request_body_type=None,
        response_body_type={
            200: GetIdentificationServiceAreaResponse,
            400: ErrorResponse,
            401: ErrorResponse,
            403: ErrorResponse,
            404: ErrorResponse,
        }
    ),
    OperationID.CreateIdentificationServiceArea: Operation(
        id="createIdentificationServiceArea",
        path="/dss/identification_service_areas/{id}",
        verb="PUT",
        request_body_type=CreateIdentificationServiceAreaParameters,
        response_body_type={
            200: PutIdentificationServiceAreaResponse,
            400: ErrorResponse,
            401: ErrorResponse,
            403: ErrorResponse,
            409: ErrorResponse,
            413: ErrorResponse,
        }
    ),
    OperationID.UpdateIdentificationServiceArea: Operation(
        id="updateIdentificationServiceArea",
        path="/dss/identification_service_areas/{id}/{version}",
        verb="PUT",
        request_body_type=UpdateIdentificationServiceAreaParameters,
        response_body_type={
            200: PutIdentificationServiceAreaResponse,
            400: ErrorResponse,
            401: ErrorResponse,
            403: ErrorResponse,
            409: ErrorResponse,
            413: ErrorResponse,
        }
    ),
    OperationID.DeleteIdentificationServiceArea: Operation(
        id="deleteIdentificationServiceArea",
        path="/dss/identification_service_areas/{id}/{version}",
        verb="DELETE",
        request_body_type=None,
        response_body_type={
            200: DeleteIdentificationServiceAreaResponse,
            400: ErrorResponse,
            401: ErrorResponse,
            403: ErrorResponse,
            404: ErrorResponse,
            409: ErrorResponse,
        }
    ),
    OperationID.SearchFlights: Operation(
        id="searchFlights",
        path="/uss/flights",
        verb="GET",
        request_body_type=None,
        response_body_type={
            200: GetFlightsResponse,
            400: ErrorResponse,
            401: ErrorResponse,
            403: ErrorResponse,
            413: ErrorResponse,
        }
    ),
    OperationID.GetFlightDetails: Operation(
        id="getFlightDetails",
        path="/uss/flights/{id}/details",
        verb="GET",
        request_body_type=None,
        response_body_type={
            200: GetFlightDetailsResponse,
            400: ErrorResponse,
            401: ErrorResponse,
            403: ErrorResponse,
            404: ErrorResponse,
        }
    ),
    OperationID.SearchSubscriptions: Operation(
        id="searchSubscriptions",
        path="/dss/subscriptions",
        verb="GET",
        request_body_type=None,
        response_body_type={
            200: SearchSubscriptionsResponse,
            400: ErrorResponse,
            401: ErrorResponse,
            403: ErrorResponse,
            413: ErrorResponse,
        }
    ),
    OperationID.GetSubscription: Operation(
        id="getSubscription",
        path="/dss/subscriptions/{id}",
        verb="GET",
        request_body_type=None,
        response_body_type={
            200: GetSubscriptionResponse,
            400: ErrorResponse,
            401: ErrorResponse,
            403: ErrorResponse,
            404: ErrorResponse,
        }
    ),
    OperationID.CreateSubscription: Operation(
        id="createSubscription",
        path="/dss/subscriptions/{id}",
        verb="PUT",
        request_body_type=CreateSubscriptionParameters,
        response_body_type={
            200: PutSubscriptionResponse,
            400: ErrorResponse,
            401: ErrorResponse,
            403: ErrorResponse,
            409: ErrorResponse,
            429: ErrorResponse,
        }
    ),
    OperationID.UpdateSubscription: Operation(
        id="updateSubscription",
        path="/dss/subscriptions/{id}/{version}",
        verb="PUT",
        request_body_type=UpdateSubscriptionParameters,
        response_body_type={
            200: PutSubscriptionResponse,
            400: ErrorResponse,
            401: ErrorResponse,
            403: ErrorResponse,
            409: ErrorResponse,
            429: ErrorResponse,
        }
    ),
    OperationID.DeleteSubscription: Operation(
        id="deleteSubscription",
        path="/dss/subscriptions/{id}/{version}",
        verb="DELETE",
        request_body_type=None,
        response_body_type={
            200: DeleteSubscriptionResponse,
            400: ErrorResponse,
            401: ErrorResponse,
            403: ErrorResponse,
            404: ErrorResponse,
            409: ErrorResponse,
        }
    ),
    OperationID.GetIdentificationServiceAreaDetails: Operation(
        id="getIdentificationServiceAreaDetails",
        path="/uss/identification_service_areas/{id}",
        verb="GET",
        request_body_type=None,
        response_body_type={
            200: GetIdentificationServiceAreaDetailsResponse,
            400: ErrorResponse,
            401: ErrorResponse,
            403: ErrorResponse,
            404: ErrorResponse,
        }
    ),
    OperationID.PostIdentificationServiceArea: Operation(
        id="postIdentificationServiceArea",
        path="/uss/identification_service_areas/{id}",
        verb="POST",
        request_body_type=PutIdentificationServiceAreaNotificationParameters,
        response_body_type={
            204: None,
            400: ErrorResponse,
            401: ErrorResponse,
            403: ErrorResponse,
            409: ErrorResponse,
        }
    ),
}



import json
import os
from datetime import timedelta
from implicitdict import ImplicitDict
from uas_standards.eurocae_ed269 import ED269Schema, ED269TimeType


def test_sample():
    with open(
        os.path.join(os.path.dirname(__file__), "test_eurocae_ed269.json")
    ) as f:
        data = json.load(f)

    ED269Schema.from_dict(data)


def test_timetype():
    class MyTimedData(ImplicitDict):
        t1: ED269TimeType
        t2: ED269TimeType
        t3: ED269TimeType
        t4: ED269TimeType

    data = ImplicitDict.parse(
        {
            "t1": "12:34:56.78Z",
            "t2": "12:34Z",
            "t3": "12:34:56.78-0100",
            "t4": "00:00:00.00+0100",
        },
        MyTimedData,
    )

    assert data["t1"].time.hour == 12
    assert data["t1"].time.minute == 34
    assert data["t1"].time.second == 56
    assert data["t1"].time.microsecond == 780000
    assert data["t1"].time.utcoffset() == timedelta(hours=0)
    assert str(data["t1"]) == "12:34:56.78Z"

    assert data["t2"].time.hour == 12
    assert data["t2"].time.minute == 34
    assert data["t2"].time.second == 0
    assert data["t2"].time.microsecond == 0
    assert data["t2"].time.utcoffset() == timedelta(hours=0)
    assert str(data["t2"]) == "12:34:00.00Z"

    assert data["t3"].time.hour == 12
    assert data["t3"].time.minute == 34
    assert data["t3"].time.second == 56
    assert data["t3"].time.microsecond == 780000
    assert data["t3"].time.utcoffset() == timedelta(hours=-1)
    assert str(data["t3"]) == "12:34:56.78-0100"

    assert data["t4"].time.hour == 0
    assert data["t4"].time.minute == 0
    assert data["t4"].time.second == 0
    assert data["t4"].time.microsecond == 0
    assert data["t4"].time.utcoffset() == timedelta(hours=1)
    assert str(data["t4"]) == "00:00:00.00+0100"

    assert data["t3"].time > data["t2"].time  # t2 is an hour earlier than t3



"""Data types and operations from Geo-Awareness Automated Test Interfaces 0.1.0 OpenAPI"""

# This file is autogenerated; do not modify manually!

from __future__ import annotations

from enum import Enum
from typing import Dict, List, Optional

from uas_standards import Operation

from implicitdict import ImplicitDict, StringBasedDateTime


UUIDv4Format = str
"""String whose format matches a version-4 UUID according to RFC 4122."""


class StatusResponseStatus(str, Enum):
    """The status of the USS automated testing interface.
    - `Starting`: the USS is starting and the automated test driver should wait before sending requests.
    - `Ready`: the USS is ready to receive test requests.
    """

    Starting = "Starting"
    Ready = "Ready"


class StatusResponse(ImplicitDict):
    status: StatusResponseStatus
    """The status of the USS automated testing interface.
    - `Starting`: the USS is starting and the automated test driver should wait before sending requests.
    - `Ready`: the USS is ready to receive test requests.
    """

    version: Optional[str]
    """Arbitrary string representing the version of the USS system to be tested."""


class GeozoneHttpsSourceFormat(str, Enum):
    """The format of the response expected from the source."""

    ED_269 = "ED-269"


class GeozoneHttpsSource(ImplicitDict):
    url: str
    """The URL at which the Geozone data shall be downloaded from."""

    format: Optional[GeozoneHttpsSourceFormat] = GeozoneHttpsSourceFormat.ED_269
    """The format of the response expected from the source."""


class GeozoneSourceResponseResult(str, Enum):
    """The status of the Geozone source and the handling of its data by the USS.
    - `Activating`: the USS is processing the request and is currently activating the Geozone data.
    - `Ready`: the Geozone data has been successfully activated and the USS is ready to receive test requests.
    - `Deactivating`: the Geozone data is being deactivated.
    - `Unsupported`: the USS cannot process the dataset type specified.
    - `Rejected`: the Geozone data was rejected because it is invalid.
    - `Error`: the Geozone data activation or deactivation failed. The message field is required in this case.
    """

    Activating = "Activating"
    Ready = "Ready"
    Deactivating = "Deactivating"
    Unsupported = "Unsupported"
    Rejected = "Rejected"
    Error = "Error"


class GeozoneSourceResponse(ImplicitDict):
    result: GeozoneSourceResponseResult
    """The status of the Geozone source and the handling of its data by the USS.
    - `Activating`: the USS is processing the request and is currently activating the Geozone data.
    - `Ready`: the Geozone data has been successfully activated and the USS is ready to receive test requests.
    - `Deactivating`: the Geozone data is being deactivated.
    - `Unsupported`: the USS cannot process the dataset type specified.
    - `Rejected`: the Geozone data was rejected because it is invalid.
    - `Error`: the Geozone data activation or deactivation failed. The message field is required in this case.
    """

    message: Optional[str]
    """Human-readable explanation of the result for debugging purpose only. This field is required when the result value is `Error`."""


class UomDimensions(str, Enum):
    M = "M"
    FT = "FT"


class VerticalReferenceType(str, Enum):
    AGL = "AGL"
    AMSL = "AMSL"


USpaceClass = str


class Restriction(str, Enum):
    PROHIBITED = "PROHIBITED"
    REQ_AUTHORISATION = "REQ_AUTHORISATION"
    CONDITIONAL = "CONDITIONAL"
    NO_RESTRICTION = "NO_RESTRICTION"


class Position(ImplicitDict):
    uomDimensions: UomDimensions

    verticalReferenceType: VerticalReferenceType

    height: Optional[float] = 0
    """Height above vertical reference datum indicated in `verticalReferenceType`, in units of `uomDimensions`."""

    longitude: Optional[float] = 0
    """Longitude, degrees east of prime meridian."""

    latitude: Optional[float] = 0
    """Latitude, degrees north of the equator."""


class ED269Filters(ImplicitDict):
    """Filter criteria for the selection of Geozones according to ED-269 characteristics."""

    uSpaceClass: Optional[USpaceClass]
    """If specified, only select Geozones which are of the specified `uSpaceClass`."""

    acceptableRestrictions: Optional[List[Restriction]]
    """If specified and non-empty, only select Geozones which are one of the specified restriction types."""


class GeozonesCheckResultGeozone(str, Enum):
    """Indication of whether one or more applicable Geozones were selected according to the selection criteria of the corresponding check.
    * Present: One or more applicable Geozones were selected. * Absent: No applicable Geozones were selected. * UnsupportedFilter: Applicable Geozones could not be selected because one or more filter criteria are not supported by the USSP.  If this value is specified, `message` must be populated. * Error: An error or condition not enumerated above occurred.  If this value is specified, `message` must be populated.
    """

    Present = "Present"
    Absent = "Absent"
    UnsupportedFilter = "UnsupportedFilter"
    Error = "Error"


class GeozonesCheckResult(ImplicitDict):
    geozone: GeozonesCheckResultGeozone
    """Indication of whether one or more applicable Geozones were selected according to the selection criteria of the corresponding check.
    * Present: One or more applicable Geozones were selected. * Absent: No applicable Geozones were selected. * UnsupportedFilter: Applicable Geozones could not be selected because one or more filter criteria are not supported by the USSP.  If this value is specified, `message` must be populated. * Error: An error or condition not enumerated above occurred.  If this value is specified, `message` must be populated.
    """

    message: Optional[str]
    """A human-readable description of why the non-standard `geozone` value was reported.  Should only be populated when appropriate according to the value of the `geozone` field."""


class CreateGeozoneSourceRequest(ImplicitDict):
    https_source: Optional[GeozoneHttpsSource]


class GeozonesFilterSet(ImplicitDict):
    """Set of filters to select only a subset of Geozones.  Only Geozones which are applicable to all specified filters within this filter set should be selected."""

    position: Optional[Position]
    """If specified, only select Geozones encompassing this position."""

    after: Optional[StringBasedDateTime]
    """If specified, only select Geozones which encompass at least some times at or after this time."""

    before: Optional[StringBasedDateTime]
    """If specified, only select Geozones which encompass at least some times at or before this time."""

    ed269: Optional[ED269Filters]


class GeozonesCheckReply(ImplicitDict):
    applicableGeozone: Optional[List[GeozonesCheckResult]] = []
    """Responses to each of the `checks` in the request.  The number of entries in this array should match the number of entries in the `checks` field of the request."""


class GeozonesCheck(ImplicitDict):
    filterSets: List[GeozonesFilterSet]
    """Select Geozones which match any of the specified filter sets."""


class GeozonesCheckRequest(ImplicitDict):
    checks: List[GeozonesCheck]


class OperationID(str, Enum):
    GetStatus = "GetStatus"
    PutGeozoneSource = "PutGeozoneSource"
    GetGeozoneSourceStatus = "GetGeozoneSourceStatus"
    DeleteGeozoneSource = "DeleteGeozoneSource"
    CheckGeozones = "CheckGeozones"


OPERATIONS: Dict[OperationID, Operation] = {
    OperationID.GetStatus: Operation(
        id="GetStatus",
        path="/status",
        verb="GET",
        request_body_type=None,
        response_body_type={
            200: StatusResponse,
            401: None,
            403: None,
            404: None,
        }
    ),
    OperationID.PutGeozoneSource: Operation(
        id="PutGeozoneSource",
        path="/geozone_sources/{geozone_source_id}",
        verb="PUT",
        request_body_type=CreateGeozoneSourceRequest,
        response_body_type={
            200: GeozoneSourceResponse,
            401: None,
            403: None,
        }
    ),
    OperationID.GetGeozoneSourceStatus: Operation(
        id="GetGeozoneSourceStatus",
        path="/geozone_sources/{geozone_source_id}",
        verb="GET",
        request_body_type=None,
        response_body_type={
            200: GeozoneSourceResponse,
            401: None,
            403: None,
            404: None,
        }
    ),
    OperationID.DeleteGeozoneSource: Operation(
        id="DeleteGeozoneSource",
        path="/geozone_sources/{geozone_source_id}",
        verb="DELETE",
        request_body_type=None,
        response_body_type={
            200: GeozoneSourceResponse,
            401: None,
            403: None,
        }
    ),
    OperationID.CheckGeozones: Operation(
        id="CheckGeozones",
        path="/geozones/check",
        verb="POST",
        request_body_type=GeozonesCheckRequest,
        response_body_type={
            200: GeozonesCheckReply,
            401: None,
            403: None,
        }
    ),
}



from enum import Enum


class Scope(str, Enum):
    StrategicCoordination = "utm.strategic_coordination"
    ConstraintManagement = "utm.constraint_management"
    ConstraintProcessing = "utm.constraint_processing"
    ConformanceMonitoringForSituationalAwareness = "utm.conformance_monitoring_sa"
    AvailabilityArbitration = "utm.availability_arbitration"


AggConfMonEvaluationFlightHours = 10
AggConfMonEvaluationPeriodDays = 7
ConflictingOIMaxUserNotificationTimeSeconds = 5
ConflictingOIMaxUSSNotificationTimeSeconds = 1
CstrPublishedNotificationLatencySeconds = 5
CstrMaxAreaKm2 = 10000
CstrMaxDeletionSeconds = 5
CstrMaxDurationHours = 24
CstrMaxPlanningHorizonDays = 56
CstrMaxTimeSendDetailsSeconds = 5
CstrMaxVertices = 1000
CstrMinEffectiveTimeBufferMinutes = 10
DSSMaxSubscriptionDurationHours = 24
ExternalDataMaxRetentionTimeHours = 24
IntersectingConstraintUserNotificationMaxSeconds = 5
IntersectionMinimumPrecisionCm = 1
MaxAggConfMonAnalysisLatencyHours = 24
MaxNonPerformanceNotificationLatencyHours = 6
MaxRecoverableTimeInNonconformingStateSeconds = 60
MaxRespondToSubscriptionNotificationSeconds = 5
MaxRespondToOIDetailsRequestSeconds = 1
OiMaxCancelTimeSeconds = 5
OiMaxDurationPerExcursionSeconds = 10
OiMaxExcursionsPerFlightHour = 18
OiMaxPlanHorizonDays = 30
OiMaxUpdateRestoreConfSeconds = 5
OiMaxUpdateTimeContingentSeconds = 5
OiMaxUpdateTimeNonconfSeconds = 5
OiMaxVertices = 10000
OiMinConformancePercent = 95
PosInfoRequestMaxResponseTimeSeconds = 5
TimeSyncMaxDifferentialSeconds = 5
TimeSyncMinPercentage = 99
TransitionToEndedMaxTimeSeconds = 5
UnableToDeliverConstraintDetailsSeconds = 5
UserOiStateChangeNotificationMaxSeconds = 5
UssFunctionFailureNotificationMaxSeconds = 5
UssOiChangeNotificationMaxSeconds = 5



"""Data types and operations from Remote ID Test Data Injection 0.1.0 OpenAPI"""

# This file is autogenerated; do not modify manually!

from __future__ import annotations

from enum import Enum
from typing import Dict, List, Optional

from uas_standards import Operation

from implicitdict import ImplicitDict, StringBasedDateTime


RIDFlightID = str
"""ID, unique to a remote ID service provider, which identifies a particular flight for which the remote ID service provider is providing remote ID services.

The following characters are not allowed: \\0 \\t \\r \\n # % / : ? @ [ \\ ]
"""


class RIDAuthData(ImplicitDict):
    """Additional authentication data."""

    format: Optional[int] = 0
    """Format of additional authentication data.

    0: None
    1: UAS ID Signature
    2: Operator ID Signature
    3: Message Set Signature
    4: Authentication Provided by Network Remote ID
    5: Specific Method
    6-9: Reserved for Spec
    10-15: Available for Private Use
    """

    data: Optional[str] = ""
    """Authentication data in form specified by `format`."""


class HorizontalAccuracy(str, Enum):
    """This is the NACp enumeration from ADS-B, plus 1m for a more complete range for UAs.

    `HAUnknown`: Unknown horizontal accuracy

    `HA10NMPlus`: > 10NM (18.52km)

    `HA10NM`: < 10NM (18.52km)

    `HA4NM`: < 4NM (7.408km)

    `HA2NM`: < 2NM (3.704km)

    `HA1NM`: < 1NM (1852m)

    `HA05NM`: < 0.5NM (926m)

    `HA03NM`: < 0.3NM (555.6m)

    `HA01NM`: < 0.1NM (185.2m)

    `HA005NM`: < 0.05NM (92.6m)

    `HA30m`: < 30m

    `HA10m`: < 10m

    `HA3m`: < 3m

    `HA1m`: < 1m
    """

    HAUnknown = "HAUnknown"
    HA10NMPlus = "HA10NMPlus"
    HA10NM = "HA10NM"
    HA4NM = "HA4NM"
    HA2NM = "HA2NM"
    HA1NM = "HA1NM"
    HA05NM = "HA05NM"
    HA03NM = "HA03NM"
    HA01NM = "HA01NM"
    HA005NM = "HA005NM"
    HA30m = "HA30m"
    HA10m = "HA10m"
    HA3m = "HA3m"
    HA1m = "HA1m"


class VerticalAccuracy(str, Enum):
    """This is the GVA enumeration from ADS-B, plus some finer values for UAs.

    `VAUnknown`: Unknown vertical accuracy

    `VA150mPlus`: > 150m

    `VA150m`: < 150m

    `VA45m`: < 45m

    `VA25m`: < 25m

    `VA10m`: < 10m

    `VA3m`: < 3m

    `VA1m`: < 1m
    """

    VAUnknown = "VAUnknown"
    VA150mPlus = "VA150mPlus"
    VA150m = "VA150m"
    VA45m = "VA45m"
    VA25m = "VA25m"
    VA10m = "VA10m"
    VA3m = "VA3m"
    VA1m = "VA1m"


class SpeedAccuracy(str, Enum):
    """This is the same enumeration scale and values from ADS-B NACv.

    `SAUnknown`: Unknown speed accuracy

    `SA10mpsPlus`: > 10 m/s

    `SA10mps`: < 10 m/s

    `SA3mps`: < 3 m/s

    `SA1mps`: < 1 m/s

    `SA03mps`: < 0.3 m/s
    """

    SAUnknown = "SAUnknown"
    SA10mpsPlus = "SA10mpsPlus"
    SA10mps = "SA10mps"
    SA3mps = "SA3mps"
    SA1mps = "SA1mps"
    SA03mps = "SA03mps"


class RIDHeightReference(str, Enum):
    """The reference datum above which the height is reported."""

    TakeoffLocation = "TakeoffLocation"
    GroundLevel = "GroundLevel"


class RIDHeight(ImplicitDict):
    """A relative altitude for the purposes of remote ID."""

    distance: float
    """Distance above reference datum.  This value is provided in meters and must have a minimum resolution of 1 meter."""

    reference: RIDHeightReference
    """The reference datum above which the height is reported."""


Latitude = float
"""Degrees of latitude north of the equator, with reference to the WGS84 ellipsoid."""


Longitude = float
"""Degrees of longitude east of the Prime Meridian, with reference to the WGS84 ellipsoid."""


class LatLngPoint(ImplicitDict):
    """Point on the earth's surface."""

    lng: Longitude

    lat: Latitude


Altitude = float
"""An altitude, in meters, above the WGS84 ellipsoid."""


class RIDOperationalStatus(str, Enum):
    """Indicates operational status of associated aircraft."""

    Undeclared = "Undeclared"
    Ground = "Ground"
    Airborne = "Airborne"
    Emergency = "Emergency"
    RemoteIDSystemFailure = "RemoteIDSystemFailure"


class RIDAircraftType(str, Enum):
    """Type of aircraft for the purposes of remote ID.

    `VTOL` is a fixed wing aircraft that can take off vertically.  `Helicopter` includes multirotor.
    """

    NotDeclared = "NotDeclared"
    Aeroplane = "Aeroplane"
    Helicopter = "Helicopter"
    Gyroplane = "Gyroplane"
    VTOL = "VTOL"
    Ornithopter = "Ornithopter"
    Glider = "Glider"
    Kite = "Kite"
    FreeBalloon = "FreeBalloon"
    CaptiveBalloon = "CaptiveBalloon"
    Airship = "Airship"
    FreeFallOrParachute = "FreeFallOrParachute"
    Rocket = "Rocket"
    TetheredPoweredAircraft = "TetheredPoweredAircraft"
    GroundObstacle = "GroundObstacle"
    Other = "Other"
    HybridLift = "HybridLift"


class UAClassificationEUCategory(str, Enum):
    EUCategoryUndefined = "EUCategoryUndefined"
    Open = "Open"
    Specific = "Specific"
    Certified = "Certified"


class UAClassificationEUClass(str, Enum):
    EUClassUndefined = "EUClassUndefined"
    Class0 = "Class0"
    Class1 = "Class1"
    Class2 = "Class2"
    Class3 = "Class3"
    Class4 = "Class4"
    Class5 = "Class5"
    Class6 = "Class6"


UAClassificationEU = dict
"""Expected keys:
* category
* class
"""



class OperatorAltitudeAltitude_type(str, Enum):
    """Source of data for the altitude field."""

    Takeoff = "Takeoff"
    Dynamic = "Dynamic"
    Fixed = "Fixed"


class OperatorAltitude(ImplicitDict):
    """Altitude associated with the Remote Pilot"""

    altitude: Optional[float]
    """Provides the Operator Altitude based on WGS-84 height above ellipsoid (HAE) (See Geodetic Altitude).  This value is provided in meters and must have a minimum resolution of 1 m."""

    altitude_type: Optional[OperatorAltitudeAltitude_type]
    """Source of data for the altitude field."""


SpecificSessionID = str
"""A unique 20 byte ID intended to identify a specific flight (session) while providing a
greater level of privacy to the operator. The first byte is the registered unique Specific Session ID
Type maintained by a registrar (see Annex A5), and the remaining 19 bytes is the Session ID per the
Specific Session ID Type specification.

Initial scheme registry entries include:

0 – reserved
1 – Internet Engineering Task Force (IETF) Drone Remote Identification Protocol (DRIP) entity ID
2 – IEEE 1609.2 HashedID8

Expressed as a hexadecimal string of the underlying data bytes.  A recipient should ignore any dashes in this value.  If fewer than 20 bytes are specified, the remaining bytes at the end are assumed to have value 0.
"""


class UASID(ImplicitDict):
    """Identification of the UAS performing this flight.  At least one field of this object must be specified."""

    serial_number: Optional[str] = ""
    """This is expressed in the CTA-2063-A Serial Number format."""

    registration_id: Optional[str] = ""
    """If a CAA provides a method of registering UAS, this number is provided by the CAA or its authorized representative.  Format is <ICAO Nationality MarkA>.<CAA Assigned ID>, ASCII encoded, only uppercase letters (A-Z), dot (.), and digits (0-9) are allowed. Example is valid for the US."""

    utm_id: Optional[str] = ""
    """A UTM-provided universally unique ID traceable to a non-obfuscated ID that acts as a "session id" to protect exposure of operationally sensitive information."""

    specific_session_id: Optional[SpecificSessionID]


class RIDAircraftPosition(ImplicitDict):
    """Position of an aircraft as reported for remote ID purposes."""

    lat: Latitude

    lng: Longitude

    alt: float
    """Geodetic altitude (NOT altitude above launch, altitude above ground, or EGM96): aircraft distance above the WGS84 ellipsoid as measured along a line that passes through the aircraft and is normal to the surface of the WGS84 ellipsoid.  This value is provided in meters and must have a minimum resolution of 1 meter."""

    accuracy_h: Optional[HorizontalAccuracy]
    """Horizontal error that is likely to be present in this reported position.  Required when `extrapolated` field is true and always in the entry for the current state."""

    accuracy_v: Optional[VerticalAccuracy]
    """Vertical error that is likely to be present in this reported position.  Required when `extrapolated` field is true and always in the entry for the current state."""

    extrapolated: Optional[bool]
    """True if this position was generated primarily by computation rather than primarily from a direct instrument measurement.  Assumed false if not specified."""

    pressure_altitude: Optional[float]
    """The uncorrected altitude (based on reference standard 29.92 inHg, 1013.25 mb) provides a reference for algorithms that utilize "altitude deltas" between aircraft.  This value is provided in meters and must have a minimum resolution of 1 meter."""


class RIDFlightDetails(ImplicitDict):
    """Details about a flight reported by a remote ID service provider.  At least one of the registration or serial fields must be filled if required by CAA."""

    id: str
    """ID for this flight, matching argument in request."""

    operator_id: Optional[str]
    """CAA-issued registration/license ID for the remote pilot or operator. """

    operator_location: Optional[LatLngPoint]
    """Location of party controlling the aircraft."""

    operation_description: Optional[str]
    """Free-text field that enables the operator to describe the purpose of a flight, if so desired."""

    auth_data: Optional[RIDAuthData]

    serial_number: Optional[str]
    """Can be specified when no registration ID exists and required by law in a region. This is expressed in the ANSI/CTA-2063-A Physical Serial Number format."""

    registration_number: Optional[str]
    """If a CAA provides a method of registering UAS, this number is provided by the CAA or its authorized representative.  Required when required by law in a region."""

    uas_id: Optional[UASID]

    operator_altitude: Optional[OperatorAltitude]

    eu_classification: Optional[UAClassificationEU]
    """When this field is specified, the Classification Type is "European Union".  If no other classification field is specified, the Classification Type is "Undeclared"."""


class RIDRecentAircraftPosition(ImplicitDict):
    time: StringBasedDateTime
    """Time at which this position applied.  RFC 3339 format, per OpenAPI specification."""

    position: RIDAircraftPosition


class RIDAircraftState(ImplicitDict):
    """State of an aircraft for the purposes of remote ID."""

    timestamp: StringBasedDateTime
    """Time at which this state was valid.  This may be the time coming from the source, such as a GPS, or the time when the system computes the values using an algorithm such as an Extended Kalman Filter (EKF).  Timestamp must be expressed with a minimum resolution of 1/10th of a second.  RFC 3339 format, per OpenAPI specification."""

    timestamp_accuracy: float
    """Declaration of timestamp accuracy, which is the largest difference between Timestamp and true time of applicability for any of the following fields: Latitude, Longitude, Geodetic Altitude, Pressure Altitude of Position, Height. to determine time of applicability of the location data provided.  Expressed in seconds, precise to 1/10ths of seconds. The accuracy reflects the 95% uncertainty bound value for the timestamp."""

    operational_status: Optional[RIDOperationalStatus]

    position: RIDAircraftPosition

    track: float
    """Direction of flight expressed as a "True North-based" ground track angle.  This value is provided in degrees East of North with a minimum resolution of 1 degree."""

    speed: float
    """Ground speed of flight in meters per second."""

    speed_accuracy: SpeedAccuracy
    """Accuracy of horizontal ground speed."""

    vertical_speed: float
    """Speed up (vertically) WGS84-HAE, m/s."""

    height: Optional[RIDHeight]

    group_radius: Optional[float]
    """Farthest horizontal distance from reported group location at which an aircraft in the group may be located (meters).  This value contains the "Operating Area Radius" data from the common data dictionary when group operation area is specified by point-radius."""

    group_ceiling: Optional[float]
    """Maximum altitude (meters WGS84-HAE) of Group Operation.  This value contains the "Operating Area Ceiling" data from the common data dictionary when group operation area is specified by point-radius."""

    group_floor: Optional[float]
    """Minimum altitude (meters WGS84-HAE) of Group Operation.  If not specified, ground level shall be assumed.  This value contains the "Operating Area Floor" data from the common data dictionary when group operation area is specified by point-radius."""

    group_count: Optional[int]
    """When operating a group (or formation or swarm), number of aircraft in group.  This value contains the "Operating Area Count" data from the common data dictionary when group operation area is specified by point-radius."""

    group_time_start: Optional[StringBasedDateTime]
    """Time at which a group operation starts.  This value contains the "Operation Area Start" data from the common data dictionary when group operation area is specified by point-radius."""

    group_time_end: Optional[StringBasedDateTime]
    """Time at which a group operation starts.  This value contains the "Operation Area End" data from the common data dictionary when group operation area is specified by point-radius."""


class TestFlightDetails(ImplicitDict):
    """The set of data with which the Service Provider system under test should respond when queried for the details of a test flight."""

    effective_after: str
    """The time after which the Service Provider system under test should respond with `details`, unless other `details` with a more recent `effective_after` time (before the current time) are available."""

    details: RIDFlightDetails
    """The details of the flight. Follows the RIDFlightDetails schema from the ASTM remote ID standard."""


class TestFlight(ImplicitDict):
    """The set of data to be injected into a Service Provider system under test for a single flight."""

    injection_id: str
    """ID of the injected test flight.  Remains the same regardless of the flight ID/UTM ID reported in the system."""

    telemetry: List[RIDAircraftState]
    """The set of telemetry data that should be injected into the system for this flight. Each element follows the RIDAircraftState schema from the ASTM remote ID standard."""

    details_responses: List[TestFlightDetails]
    """The details of the flight as a function of time."""


class CreateTestParameters(ImplicitDict):
    """A complete set of data to be injected into a Service Provider system under test."""

    requested_flights: List[TestFlight]
    """One or more logical flights, each containing test data to inject into the system. Elements should be sorted in ascending order of `timestamp`."""


class ChangeTestResponse(ImplicitDict):
    injected_flights: List[TestFlight]
    """The complete set of test data actually injected into the Service Provider system under test."""

    version: str
    """Version of test.  Used to delete test."""


class DeleteTestResponse(ImplicitDict):
    injected_flights: List[TestFlight]
    """The complete set of test data deleted."""


class OperationID(str, Enum):
    CreateTest = "createTest"
    DeleteTest = "deleteTest"


OPERATIONS: Dict[OperationID, Operation] = {
    OperationID.CreateTest: Operation(
        id="createTest",
        path="/tests/{test_id}",
        verb="PUT",
        request_body_type=CreateTestParameters,
        response_body_type={
            200: ChangeTestResponse,
            409: None,
        }
    ),
    OperationID.DeleteTest: Operation(
        id="deleteTest",
        path="/tests/{test_id}/{version}",
        verb="DELETE",
        request_body_type=None,
        response_body_type={
            200: DeleteTestResponse,
        }
    ),
}
